"""
The folder is called "ode" but this
module is "ivp" because in the future,
there might be more ode-based problems,
such as bvp.
"""

import numpy as np

from probnum.diffeq.ode.ode import ODE


__all__ = ["logistic", "fitzhughnagumo", "lotkavolterra", "IVP"]


def logistic(timespan, initrv, params=(3.0, 1.0)):
    """
    Initial value problem (IVP) based on the logistic ODE.

    The logistic ODE is defined through

    .. math:: f(t, x) = a  x  \\left(1 - \\frac{x}{b}\\right)

    for some parameters :math:`(a, b)`.
    Default is :math:`(a, b)=(3.0, 1.0)`. This implementation includes
    the Jacobian :math:`J_f` of :math:`f` as well as a closed form
    solution given by

    .. math:: f(t) = \\frac{b x_0 \\exp(a t)}{b + x_0 [\\exp(at) - 1]}

    where :math:`x_0= x(t_0)` is the initial value.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        RandomVariable that  describes the belief over the initial
        value. Usually its distribution is Dirac (noise-free)
        or Normal (noisy). To replicate "classical" initial values
        use the Dirac distribution.
    params : (float, float), optional
        Parameters :math:`(a, b)` for the logistic IVP.
        Default is :math:`(a, b) = (3.0, 1.0)`.

    Returns
    -------
    IVP
        IVP object describing the logistic IVP with the prescribed
        configuration.
    """

    def rhs(t, x):
        return log_rhs(t, x, params)

    def jac(t, x):
        return log_jac(t, x, params)

    def hess(t, x):
        return log_hess(t, x, params)

    def sol(t):
        return log_sol(t, params, initrv.mean())

    return IVP(timespan, initrv, rhs, jac, hess, sol)


def log_rhs(t, x, params):
    """RHS for logistic model."""
    l0, l1 = params
    return l0 * x * (1.0 - x / l1)


def log_jac(t, x, params):
    """Jacobian for logistic model."""
    l0, l1 = params
    return np.array([l0 - l0 / l1 * 2 * x])


def log_hess(t, x, params):
    """Hessian for logistic model."""
    l0, l1 = params
    return np.array([[-2 * l0 / l1]])


def log_sol(t, params, x0):
    """Solution for logistic model."""
    l0, l1 = params
    nomin = l1 * x0 * np.exp(l0 * t)
    denom = l1 + x0 * (np.exp(l0 * t) - 1)
    return nomin / denom


def fitzhughnagumo(timespan, initrv, params=(0.0, 0.08, 0.07, 1.25)):
    """
    Initial value problem (IVP) based on the FitzHugh-Nagumo model.

    The FitzHugh-Nagumo (FHN) model is defined through

    .. math:: f(t, x) =
        \\begin{pmatrix} x_1 - \\frac{1}{3}x_1^3 - x_2 + a \\\\
        \\frac{1}{d} (x_1 + b - c x_2)  \\end{pmatrix}

    for some parameters :math:`(a, b, c, d)`.
    Default is :math:`(a, b)=(0.0, 0.08, 0.07, 1.25)`.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        RandomVariable that  describes the belief over the initial
        value. Usually its distribution is Dirac (noise-free)
        or Normal (noisy). To replicate "classical" initial values
        use the Dirac distribution.
    params : (float, float, float, float), optional
        Parameters :math:`(a, b, c, d)` for the logistic IVP.
        Default is :math:`(a, b, c, d)=(0.0, 0.08, 0.07, 1.25)`.

    Returns
    -------
    IVP
        IVP object describing the logistic IVP with the prescribed
        configuration.
    """

    def rhs(t, x):
        return fhn_rhs(t, x, params)

    def jac(t, x):
        return fhn_jac(t, x, params)

    return IVP(timespan, initrv, rhs, jac)


def fhn_rhs(t, x, params):
    """RHS for FitzHugh-Nagumo model."""
    x1, x2 = x
    a, b, c, d = params
    return np.array([x1 - x1 ** 3 / 3 - x2 + a, (x1 + b - c * x2) / d])


def fhn_jac(t, x, params):
    """Jacobian for FitzHugh-Nagumo model."""
    x1, x2 = x
    a, b, c, d = params
    return np.array([[1 - x1 ** 2, -1], [1.0 / d, -c / d]])


def lotkavolterra(timespan, initrv, params=(0.5, 0.05, 0.5, 0.05)):
    """
    Initial value problem (IVP) based on the Lotka-Volterra model.

    The Lotka-Volterra (LV) model is defined through

    .. math:: f(t, x) =
        \\begin{pmatrix} a x_1 - bx_1x_2 \\\\ -c x_2 + d x_1 x_2
        \\end{pmatrix}

    for some parameters :math:`(a, b, c, d)`.
    Default is :math:`(a, b)=(0.5, 0.05, 0.5, 0.05)`.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        RandomVariable that  describes the belief over the initial
        value. Usually its distribution is Dirac (noise-free)
        or Normal (noisy). To replicate "classical" initial values
        use the Dirac distribution.
    params : (float, float, float, float), optional
        Parameters :math:`(a, b, c, d)` for the logistic IVP.
        Default is :math:`(a, b, c, d)=(0.5, 0.05, 0.5, 0.05)`.

    Returns
    -------
    IVP
        IVP object describing the logistic IVP with the prescribed
        configuration.
    """

    def rhs(t, x):
        return lv_rhs(t, x, params)

    def jac(t, x):
        return lv_jac(t, x, params)

    return IVP(timespan, initrv, rhs, jac)


def lv_rhs(t, x, params):
    """RHS for Lotka-Volterra"""
    a, b, c, d = params
    x1, x2 = x
    return np.array([a * x1 - b * x1 * x2, -c * x2 + d * x1 * x2])


def lv_jac(t, x, params):
    """Jacobian for Lotka-Volterra"""
    a, b, c, d = params
    x1, x2 = x
    return np.array([[a - b * x2, -b * x1], [d * x2, -c + d * x1]])


class IVP(ODE):
    """
    Initial value problems (IVP).

    This class descibes initial value problems based on systems of
    first order ordinary differential equations (ODEs),

    .. math:: \\dot x(t) = f(t, x(t)), \\quad x(t_0) = x_0,
        \\quad t \\in [t_0, T]

    It provides options for defining custom right-hand side (RHS)
    functions, their Jacobians and closed form solutions.

    Since we use them for probabilistic ODE solvers these functions
    fit into the probabilistic framework as well. That is,
    the initial value is a RandomVariable object with some
    distribution that reflects the prior belief over the initial
    value. To recover "classical" initial values one can use the
    Dirac distribution.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        RandomVariable that  describes the belief over the initial
        value. Usually its distribution is Dirac (noise-free)
        or Normal (noisy). To replicate "classical" initial values
        use the Dirac distribution.
        Implementation depends on the mean of this RandomVariable,
        so please only use RandomVariable objects with available
        means, e.g. Diracs or Normals.
    rhs : callable, signature: ``(t, x, **kwargs)``
        RHS function
        :math:`f : [0, T] \\times \\mathbb{R}^d \\rightarrow \\mathbb{R}^d`
        of the ODE system. As such it takes a float and an
        np.ndarray of shape (d,) and returns a np.ndarray
        of shape (d,). As of now, no vectorization is supported
        (nor needed).
    jac : callable, signature: ``(t, x, **kwargs)``, optional
        Jacobian of RHS function
        :math:`J_f : [0, T] \\times \\mathbb{R}^d \\rightarrow \\mathbb{R}^d`
        of the ODE system. As such it takes a float and an
        np.ndarray of shape (d,) and returns a np.ndarray
        of shape (d,). As of now, no vectorization is supported
        (nor needed).
    sol : callable, signature: ``(t, **kwargs)``, optional
        Solution of IVP.

    See Also
    --------
    ODE : Abstract interface for  ordinary differential equations.

    Examples
    --------
    >>> from probnum.diffeq import IVP
    >>> rhsfun = lambda t, x, **kwargs: 2.0*x
    >>> from probnum.prob import Dirac, Normal, RandomVariable
    >>> initrv = RandomVariable(distribution=Dirac(0.1))
    >>> timespan = (0, 10)
    >>> ivp = IVP(timespan, initrv, rhsfun)
    >>> print(ivp.rhs(0., 2.))
    4.0
    >>> print(ivp.timespan)
    [0, 10]
    >>> print(ivp.t0)
    0

    >>> initrv = RandomVariable(distribution=Normal(0.1, 1.0))
    >>> ivp = IVP(timespan, initrv, rhsfun)
    >>> jac = lambda t, x, **kwargs: 2.0
    >>> ivp = IVP(timespan, initrv, rhs=rhsfun, jac=jac)
    >>> print(ivp.rhs(0., 2.))
    4.0
    >>> print(ivp.jacobian(100., -1))
    2.0
    """

    def __init__(self, timespan, initrv, rhs, jac=None, hess=None, sol=None):
        """
        """
        self.initrv = initrv
        super().__init__(timespan=timespan, rhs=rhs, jac=jac, hess=hess, sol=sol)

    @property
    def initialdistribution(self):
        """
        Distribution of the initial random variable.
        """
        return self.initrv

    @property
    def initialrandomvariable(self):
        """
        Initial random variable.
        """
        return self.initrv

    @property
    def ndim(self):
        """
        Spatial dimension of the IVP problem.

        Depends on the mean of the initial random variable.
        """
        if np.isscalar(self.initrv.mean()):
            return 1
        else:
            return len(self.initrv.mean())



"""
Probabilistic numerical methods for solving linear systems.

This module provides routines to solve linear systems of equations in a Bayesian framework. This means that a prior
distribution over elements of the linear system can be provided and is updated with information collected by the solvers
to return a posterior distribution.
"""

import warnings

import numpy as np
import scipy.sparse

from probnum import prob
from probnum.linalg import linops
from probnum import utils
from probnum.linalg.linearsolvers.matrixbased import (
    AsymmetricMatrixBasedSolver,
    NoisySymmetricMatrixBasedSolver,
    SymmetricMatrixBasedSolver,
)
from probnum.linalg.linearsolvers.solutionbased import SolutionBasedSolver


def problinsolve(
    A,
    b,
    A0=None,
    Ainv0=None,
    x0=None,
    assume_A="sympos",
    maxiter=None,
    atol=10 ** -6,
    rtol=10 ** -6,
    callback=None,
    **kwargs
):
    """
    Infer a solution to the linear system :math:`A x = b` in a Bayesian framework.

    Probabilistic linear solvers infer solutions to problems of the form

    .. math:: Ax=b,

    where :math:`A \\in \\mathbb{R}^{n \\times n}` and :math:`b \\in \\mathbb{R}^{n}`. They return a probability measure
    which quantifies uncertainty in the output arising from finite computational resources. This solver can take prior
    information either on the linear operator :math:`A` or its inverse :math:`H=A^{-1}` in
    the form of a random variable ``A0`` or ``Ainv0`` and outputs a posterior belief over :math:`A` or :math:`H`. This
    code implements the method described in Wenger et al. [1]_ based on the work in Hennig et al. [2]_.

    Parameters
    ----------
    A : array-like or LinearOperator, shape=(n,n)
        A square linear operator (or matrix). Only matrix-vector products :math:`Av` are used internally.
    b : array_like or RandomVariable, shape=(n,) or (n, nrhs)
        Right-hand side vector, matrix or random variable in :math:`A x = b`. For multiple right hand sides, ``nrhs``
        problems are solved sequentially with the posteriors over the matrices acting as priors for subsequent solves.
        If the right-hand-side is assumed to be noisy, every iteration of the solver samples a realization from ``b``.
    A0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior belief over the linear operator
        :math:`A`. If an array or linear operator is given, a prior distribution is chosen automatically.
    Ainv0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior belief over the inverse
        :math:`H=A^{-1}`. This can be viewed as taking the form of a pre-conditioner. If an array or linear operator is
        given, a prior distribution is chosen automatically.
    x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
        Optional. Prior belief for the solution of the linear system. Will be ignored if ``Ainv0`` is given.
    assume_A : str, default="sympos"
        Assumptions on the linear operator which can influence solver choice and behavior. The available options are
        (combinations of)

        ====================  =========
         generic matrix       ``gen``
         symmetric            ``sym``
         positive definite    ``pos``
         (additive) noise     ``noise``
        ====================  =========

    maxiter : int, optional
        Maximum number of iterations. Defaults to :math:`10n`, where :math:`n` is the dimension of :math:`A`.
    atol : float, optional
        Absolute convergence tolerance.
    rtol : float, optional
        Relative convergence tolerance.
    callback : function, optional
        User-supplied function called after each iteration of the linear solver. It is called as
        ``callback(xk, Ak, Ainvk, sk, yk, alphak, resid, **kwargs)`` and can be used to return quantities from the
        iteration. Note that depending on the function supplied, this can slow down the solver considerably.
    kwargs : optional
        Optional keyword arguments passed onto the solver iteration.

    Returns
    -------
    x : RandomVariable, shape=(n,) or (n, nrhs)
        Approximate solution :math:`x` to the linear system. Shape of the return matches the shape of ``b``.
    A : RandomVariable, shape=(n,n)
        Posterior belief over the linear operator.
    Ainv : RandomVariable, shape=(n,n)
        Posterior belief over the linear operator inverse :math:`H=A^{-1}`.
    info : dict
        Information on convergence of the solver.

    Raises
    ------
    ValueError
        If size mismatches detected or input matrices are not square.
    LinAlgError
        If the matrix ``A`` is singular.
    LinAlgWarning
        If an ill-conditioned input ``A`` is detected.

    Notes
    -----
    For a specific class of priors the posterior mean of :math:`x_k=Hb` coincides with the iterates of the conjugate
    gradient method. The matrix-based view taken here recovers the solution-based inference of :func:`bayescg` [3]_.

    References
    ----------
    .. [1] Wenger, J. and Hennig, P., Probabilistic Linear Solvers for Machine Learning, 2020
    .. [2] Hennig, P., Probabilistic Interpretation of Linear Solvers, *SIAM Journal on Optimization*, 2015, 25, 234-260
    .. [3] Bartels, S. et al., Probabilistic Linear Solvers: A Unifying View, *Statistics and Computing*, 2019

    See Also
    --------
    bayescg : Solve linear systems with prior information on the solution.

    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(1)
    >>> n = 20
    >>> A = np.random.rand(n, n)
    >>> A = 0.5 * (A + A.T) + 5 * np.eye(n)
    >>> b = np.random.rand(n)
    >>> x, A, Ainv, info = problinsolve(A=A, b=b)
    >>> print(info["iter"])
    9
    """

    # Check linear system for type and dimension mismatch
    _check_linear_system(A=A, b=b, A0=A0, Ainv0=Ainv0, x0=x0)

    # Check matrix assumptions for correctness
    assume_A = assume_A.lower()
    _assume_A_tmp = assume_A
    for allowed_str in ["gen", "sym", "pos", "noise"]:
        _assume_A_tmp = _assume_A_tmp.replace(allowed_str, "")
    if _assume_A_tmp != "":
        raise ValueError(
            "Assumption '{}' contains unrecognized linear operator properties.".format(
                assume_A
            )
        )

    # Transform the linear system to an appropriate form
    A, b, x0 = _preprocess_linear_system(A=A, b=b, x0=x0)

    # Parameter initialization
    n = A.shape[0]
    nrhs = b.shape[1]
    x = x0
    info = {}

    # Set convergence parameters
    if maxiter is None:
        maxiter = n * 10

    if nrhs > 1:
        # Iteratively solve for multiple right hand sides (with posteriors as new priors)
        for i in range(nrhs):
            if i > 0:
                x = None  # Only use prior information on Ainv for multiple rhs
            # Select and initialize solver
            linear_solver = _init_solver(
                A=A,
                b=utils.as_colvec(b[:, i]),
                A0=A0,
                Ainv0=Ainv0,
                x0=x,
                assume_A=assume_A,
            )

            # Solve linear system
            x, A0, Ainv0, info = linear_solver.solve(
                maxiter=maxiter, atol=atol, rtol=rtol, callback=callback, **kwargs
            )

        # Return Ainv @ b for multiple rhs
        x = Ainv0 @ b
    else:
        # Single right hand side
        linear_solver = _init_solver(
            A=A, b=b, A0=A0, Ainv0=Ainv0, x0=x, assume_A=assume_A
        )

        # Solve linear system
        x, A0, Ainv0, info = linear_solver.solve(
            maxiter=maxiter, atol=atol, rtol=rtol, callback=callback, **kwargs
        )

    # Check result and issue warnings (e.g. singular or ill-conditioned matrix)
    _postprocess(info=info, A=A)

    return x, A0, Ainv0, info


def bayescg(A, b, x0=None, maxiter=None, atol=None, rtol=None, callback=None):
    """
    Conjugate Gradients using prior information on the solution of the linear system.

    In the setting where :math:`A` is a symmetric positive-definite matrix, this solver takes prior information
    on the solution and outputs a posterior belief over :math:`x`. This code implements the
    method described in Cockayne et al. [1]_.

    Note that the solution-based view of BayesCG and the matrix-based view of :meth:`problinsolve` correspond [2]_.

    Parameters
    ----------
    A : array-like or LinearOperator, shape=(n,n)
        A square linear operator (or matrix). Only matrix-vector products :math:`Av` are used internally.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.
    x0 : array-like or RandomVariable, shape=(n,) or or (n, nrhs)
        Prior belief over the solution of the linear system.
    maxiter : int
        Maximum number of iterations. Defaults to :math:`10n`, where :math:`n` is the dimension of :math:`A`.
    atol : float, optional
        Absolute residual tolerance. If :math:`\\lVert r_i \\rVert = \\lVert Ax_i - b \\rVert < \\text{atol}`, the
        iteration terminates.
    rtol : float, optional
        Relative residual tolerance. If :math:`\\lVert r_i \\rVert  < \\text{rtol} \\lVert b \\rVert`, the
        iteration terminates.
    callback : function, optional
        User-supplied function called after each iteration of the linear solver. It is called as
        ``callback(xk, sk, yk, alphak, resid, **kwargs)`` and can be used to return quantities from the iteration. Note
        that depending on the function supplied, this can slow down the solver.

    References
    ----------
    .. [1] Cockayne, J. et al., A Bayesian Conjugate Gradient Method, *Bayesian Analysis*, 2019, 14, 937-1012
    .. [2] Bartels, S. et al., Probabilistic Linear Solvers: A Unifying View, *Statistics and Computing*, 2019

    See Also
    --------
    problinsolve : Solve linear systems in a Bayesian framework.
    """
    # Check linear system for type and dimension mismatch
    _check_linear_system(A=A, b=b, x0=x0)

    # Preprocess linear system
    A, b, x0 = _preprocess_linear_system(A=A, b=b, x0=x0)

    # Set default convergence parameters
    n = A.shape[0]
    if maxiter is None:
        maxiter = n * 10

    # Solve linear system
    x, info = SolutionBasedSolver(A=A, b=b, x0=x0).solve(
        callback=callback, maxiter=maxiter, atol=atol, rtol=rtol
    )

    # Check result and issue warnings (e.g. singular or ill-conditioned matrix)
    _postprocess(info=info, A=A)

    return x, info


def _check_linear_system(A, b, A0=None, Ainv0=None, x0=None):
    """
    Check linear system compatibility.

    Raises an exception if the input arguments are not of the right type or not compatible.

    Parameters
    ----------
    A : array-like or LinearOperator, shape=(n,n)
        A square linear operator (or matrix). Only matrix-vector products :math:`Av` are used internally.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.
    A0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior belief over the linear operator
        :math:`A`.
    Ainv0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior belief over the inverse
        :math:`H=A^{-1}`.
    x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
        Optional. Prior belief for the solution of the linear system. Will be ignored if ``Ainv0`` is given.

    Raises
    ------
    ValueError
        If type or size mismatches detected or inputs ``A`` and ``Ainv`` are not square.
    """
    # Check types
    linop_types = (
        np.ndarray,
        scipy.sparse.spmatrix,
        scipy.sparse.linalg.LinearOperator,
        prob.RandomVariable,
    )
    vector_types = (np.ndarray, scipy.sparse.spmatrix, prob.RandomVariable)
    if not isinstance(A, linop_types):
        raise ValueError(
            "A must be either an array, a linear operator or a random variable."
        )
    if not isinstance(b, vector_types):
        raise ValueError(
            "The right hand side must be a (sparse) array or a random variable."
        )
    if A0 is not None and not isinstance(A0, prob.RandomVariable):
        raise ValueError("The prior belief over A must be a random variable.")
    if Ainv0 is not None and not isinstance(Ainv0, linop_types):
        raise ValueError(
            "The inverse of A must be either an array, a linear operator or a random variable of either."
        )
    if x0 is not None and not isinstance(x0, vector_types):
        raise ValueError("The initial guess for the solution must be a (sparse) array.")

    # Prior distribution mismatch
    if (
        isinstance(A0, prob.RandomVariable) or isinstance(Ainv0, prob.RandomVariable)
    ) and isinstance(x0, prob.RandomVariable):
        raise ValueError(
            "Cannot specify distributions on the linear operator and the solution simultaneously."
        )

    # Dimension mismatch
    if A.shape[0] != b.shape[0]:
        raise ValueError("Dimension mismatch. The dimensions of A and b must match.")
    if Ainv0 is not None and A.shape != Ainv0.shape:
        raise ValueError(
            "Dimension mismatch. The dimensions of A and Ainv0 must match."
        )
    if A0 is not None and A.shape != A0.shape:
        raise ValueError("Dimension mismatch. The dimensions of A and A0 must match.")
    if x0 is not None and A.shape[1] != x0.shape[0]:
        raise ValueError("Dimension mismatch. The dimensions of A and x0 must match.")

    # Square matrices
    if A.shape[0] != A.shape[1]:
        raise ValueError("Matrix A must be square.")
    if A0 is not None and A0.shape[0] != A0.shape[1]:
        raise ValueError("A0 must be square.")
    if Ainv0 is not None and Ainv0.shape[0] != Ainv0.shape[1]:
        raise ValueError("The inverse of A must be square.")


def _preprocess_linear_system(A, b, x0=None):
    """
    Transform the linear system to an appropriate form.

    Parameters
    ----------
    A : array-like or LinearOperator, shape=(n,n)
        A square linear operator (or matrix). Only matrix-vector products :math:`Av` are used internally.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.
    x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
        Optional. Prior belief for the solution of the linear system. Will be ignored if ``Ainv0`` is given.

    Returns
    -------
    A : RandomVariable, shape=(n,n)
        Prior belief over the linear operator :math:`A`.
    b : array-like, shape=(n,) or (n, nrhs)
        Right-hand-side of the linear system.
    x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
        Optional. Prior belief for the solution of the linear system. Will be ignored if ``Ainv0`` is given.
    """
    # Transform linear system to correct dimensions
    if not isinstance(b, prob.RandomVariable):
        b = utils.as_colvec(b)  # (n,) -> (n, 1)
    if x0 is not None:
        x0 = utils.as_colvec(x0)  # (n,) -> (n, 1)

    return A, b, x0


def _init_solver(A, b, A0, Ainv0, x0, assume_A):
    """
    Selects and initializes an appropriate instance of the probabilistic linear solver based on the system properties
    and prior information given.

    Parameters
    ----------
    A : array-like or LinearOperator, shape=(n,n)
        A square linear operator (or matrix). Only matrix-vector products :math:`Av` are used internally.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.
    A0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior belief over the linear operator
        :math:`A`.
    Ainv0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior belief over the inverse
        :math:`H=A^{-1}`.
    x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
        Optional. Prior belief for the solution of the linear system. Will be ignored if ``Ainv0`` is given.
    assume_A : str
        Assumptions on the linear operator, which can influence solver choice or behavior. The available options are
        (combinations of)

        ====================  =========
         generic matrix       ``gen``
         symmetric            ``sym``
         positive definite    ``pos``
         (additive) noise     ``noise``
        ====================  =========

    Returns
    -------
    linear_solver : ProbabilisticLinearSolver
        A type of probabilistic linear solver implementing the solve method for linear systems.

    """
    # Choose matrix based view if not clear from arguments
    if (Ainv0 is not None or A0 is not None) and isinstance(x0, prob.RandomVariable):
        warnings.warn(
            "Cannot use prior uncertainty on both the matrix (inverse) and the solution. The latter will be ignored."
        )
        x0 = x0.mean()

    # Extract information from priors
    # System matrix is symmetric
    if isinstance(A0, prob.RandomVariable):
        if isinstance(A0.cov(), linops.SymmetricKronecker) and "sym" not in assume_A:
            assume_A += "sym"
    if isinstance(Ainv0, prob.RandomVariable):
        if isinstance(Ainv0.cov(), linops.SymmetricKronecker) and "sym" not in assume_A:
            assume_A += "sym"
    # System matrix is NOT stochastic
    if (
        not isinstance(A, prob.RandomVariable)
        and not isinstance(A, scipy.sparse.linalg.LinearOperator)
        and "noise" in assume_A
    ):
        warnings.warn(
            "A is assumed to be noisy, but is neither a random variable nor a linear operator. Use exact "
            "probabilistic linear solver instead."
        )

    # Solution-based view
    if isinstance(x0, prob.RandomVariable):
        return SolutionBasedSolver(A=A, b=b, x0=x0)
    # Matrix-based view
    else:
        if "sym" in assume_A and "pos" in assume_A:
            if "noise" in assume_A:
                return NoisySymmetricMatrixBasedSolver(
                    A=A, b=b, x0=x0, A0=A0, Ainv0=Ainv0
                )
            else:
                return SymmetricMatrixBasedSolver(A=A, b=b, x0=x0, A0=A0, Ainv0=Ainv0)
        elif "sym" not in assume_A and "pos" in assume_A:
            return AsymmetricMatrixBasedSolver(A=A, b=b, x0=x0)
        else:
            raise NotImplementedError


def _postprocess(info, A):
    """
    Postprocess the linear system and its solution.

    Raises exceptions or warnings based on the properties of the linear system and the solver iteration.

    Parameters
    ----------
    info : dict
        Convergence information output by a probabilistic linear solver.
    A : array-like or LinearOperator, shape=(n,n)
        A square linear operator (or matrix).

    Raises
    ------
    LinAlgError
        If the matrix ``A`` is singular.
    LinAlgWarning
        If an ill-conditioned system matrix ``A`` is detected.
    """
    rel_cond = info["rel_cond"]

    # Get the correct machine epsilon for the precision used.
    # if A.dtype.char in 'fF':  # single precision
    #     lamch = scipy.linalg.get_lapack_funcs('lamch', dtype='f')
    # else:
    #     lamch = scipy.linalg.get_lapack_funcs('lamch', dtype='d')
    # machine_eps = lamch('E')
    machine_eps = 10 ** -16

    # Singular matrix
    # # TODO: get info from solver
    # if False:
    #     raise scipy.linalg.LinAlgError("The system matrix A is singular.")
    # Ill-conditioned matrix A
    if rel_cond is not None and 1 / rel_cond < machine_eps:
        warnings.warn(
            "Ill-conditioned matrix detected (estimated rcond={:.6g}). Results are likely inaccurate.".format(
                rel_cond
            ),
            scipy.linalg.LinAlgWarning,
            stacklevel=3,
        )



"""
Abstract ODESolver class. Interface for Runge-Kutta, ODEFilter.
"""

from abc import ABC, abstractmethod


class ODESolver(ABC):
    """
    Interface for ODESolver.
    """

    def __init__(self, steprule):
        """
        An ODESolver is an object governed by a stepsize rule.
        That is: constant steps or adaptive steps.
        """
        self.steprule = steprule

    @abstractmethod
    def solve(self, ivp, minstep, maxstep, **kwargs):
        """
        Every ODE solver has a solve() method.
        Optional: callback function. Allows  e.g. printing variables
        at runtime.
        """
        raise NotImplementedError



"""
Solution-based probabilistic linear solvers.

Implementations of solution-based linear solvers which perform inference on the solution of a linear system given linear
observations.
"""

import warnings
import numpy as np

from probnum.linalg.linearsolvers.matrixbased import ProbabilisticLinearSolver


class SolutionBasedSolver(ProbabilisticLinearSolver):
    """
    Solver iteration of BayesCG.

    Implements the solve iteration of the solution-based solver BayesCG [1]_.

    Parameters
    ----------
    A : array-like or LinearOperator or RandomVariable, shape=(n,n)
        The square matrix or linear operator of the linear system.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.

    References
    ----------
    .. [1] Cockayne, J. et al., A Bayesian Conjugate Gradient Method, *Bayesian Analysis*, 2019, 14, 937-1012
    """

    def __init__(self, A, b, x0=None):
        self.x0 = x0
        super().__init__(A=A, b=b)

    def has_converged(self, iter, maxiter, resid=None, atol=None, rtol=None):
        """
        Check convergence of a linear solver.

        Evaluates a set of convergence criteria based on its input arguments to decide whether the iteration has converged.

        Parameters
        ----------
        iter : int
            Current iteration of solver.
        maxiter : int
            Maximum number of iterations
        resid : array-like
            Residual vector :math:`\\lVert r_i \\rVert = \\lVert Ax_i - b \\rVert` of the current iteration.
        atol : float
            Absolute residual tolerance. Stops if :math:`\\lVert r_i \\rVert < \\text{atol}`.
        rtol : float
            Relative residual tolerance. Stops if :math:`\\lVert r_i \\rVert < \\text{rtol} \\lVert b \\rVert`.

        Returns
        -------
        has_converged : bool
            True if the method has converged.
        convergence_criterion : str
            Convergence criterion which caused termination.
        """
        # maximum iterations
        if iter >= maxiter:
            warnings.warn(
                message="Iteration terminated. Solver reached the maximum number of iterations."
            )
            return True, "maxiter"
        # residual below error tolerance
        elif np.linalg.norm(resid) <= atol:
            return True, "resid_atol"
        elif np.linalg.norm(resid) <= rtol * np.linalg.norm(self.b):
            return True, "resid_rtol"
        else:
            return False, ""

    def solve(self, callback=None, maxiter=None, atol=None, rtol=None):
        raise NotImplementedError



"""
"""

from abc import ABC, abstractmethod

__all__ = ["StepRule", "ConstantSteps", "AdaptiveSteps"]


class StepRule(ABC):
    """
    (Adaptive) step size rules for ODE solvers.
    """

    @abstractmethod
    def suggest(self, laststep, errorest, **kwargs):
        """
        Suggests a new step h_{n+1} given error estimate
        e_n at step h_n.
        """
        raise NotImplementedError

    @abstractmethod
    def is_accepted(self, proposedstep, errorest, **kwargs):
        """
        Checks if proposed step should be accepted or not.
        Variable "proposedstep" not used yet, but may be
        important in the future, e.g. if we decide that
        instead of tol_per_step (see AdaptiveSteps) we want to be able to
        control tol_per_unitstep.
        """
        raise NotImplementedError


class ConstantSteps(StepRule):
    """
    Constant step size rule for ODE solvers.
    """

    def __init__(self, stepsize):
        """
        """
        self.step = stepsize

    def suggest(self, laststep, errorest, **kwargs):
        """
        """
        return self.step

    def is_accepted(self, proposedstep, errorest, **kwargs):
        """
        Meaningless since always True.
        """
        return True


class AdaptiveSteps(StepRule):
    """
    Adaptive step size selection based on
    tolerance per step.

    By default, there is no being "too small" for a step. However, a
    Warning is printed if the suggested step is smaller than roughly
    machine precision: 1e-15.

    Parameters
    ----------
    tol_per_step : float
        Tolerance per step (absolute)
    localconvrate : float
        (Estimated) convergence rate of the solver.
    limitchange : list with 2 elements, optional
        Lower and upper bounds for computed change of step.
    safetyscale : float, optional
        Safety factor for proposal of distributions, 0 << safetyscale < 1
    """

    def __init__(
        self, tol_per_step, localconvrate, limitchange=(0.1, 5.0), safetyscale=0.95
    ):
        self.tol_per_step = float(tol_per_step)
        self.safetyscale = float(safetyscale)
        self.localconvrate = float(localconvrate + 1)
        self.limitchange = limitchange

    def suggest(self, laststep, errorest, **kwargs):
        """
        """
        small, large = self.limitchange
        ratio = self.tol_per_step / (laststep * errorest)
        change = self.safetyscale * ratio ** (1.0 / self.localconvrate)
        if change < small:
            step = small * laststep
        elif large < change:
            step = large * laststep
        else:
            step = change * laststep
        if step < 1e-15:
            print("Warning: Stepsize is num. zero (h=%.1e)" % step)
        return step

    def is_accepted(self, proposedstep, errorest, **kwargs):
        """
        """
        if errorest * proposedstep < self.tol_per_step:
            return True
        else:
            return False



"""
Random Variables.

This module implements random variables. Random variables are the main in- and outputs of probabilistic numerical
methods.
"""

import operator

import numpy as np
import scipy.stats
import scipy.sparse
import scipy._lib._util

from probnum.prob.distributions.distribution import Distribution
from probnum.prob.distributions.dirac import Dirac
from probnum.prob.distributions.normal import Normal


class RandomVariable:
    """
    Random variables are the main objects used by probabilistic numerical methods.

    Every probabilistic numerical method takes a random variable encoding the prior distribution as input and outputs a
    random variable whose distribution encodes the uncertainty arising from finite computation. The generic signature
    of a probabilistic numerical method is:

    ``output_rv = probnum_method(input_rv, method_params)``

    In practice, most random variables used by methods in ProbNum have Dirac or Gaussian measure.

    Instances of :class:`RandomVariable` can be added, multiplied, etc. with arrays and linear operators. This may
    change their ``distribution`` and not necessarily all previously available methods are retained.

    Parameters
    ----------
    shape : tuple
        Shape of realizations of this random variable.
    dtype : numpy.dtype or object
        Data type of realizations of this random variable. If ``object`` will be converted to ``numpy.dtype``.
    distribution : Distribution
        Probability distribution of the random variable.

    See Also
    --------
    asrandvar : Transform into a :class:`RandomVariable`.
    Distribution : A class representing probability distributions.

    Examples
    --------
    """

    def __init__(self, shape=None, dtype=None, distribution=None):
        """Create a new random variable."""
        self._set_distribution(distribution)
        self._set_dtype(distribution, dtype)
        self._set_shape(distribution, shape)

    def _set_distribution(self, distribution):
        """
        Set distribution of random variable.
        """
        if isinstance(distribution, Distribution):
            self._distribution = distribution
        elif distribution is None:
            self._distribution = Distribution()
        else:
            raise ValueError(
                "The distribution parameter must be an " "instance of `Distribution`."
            )

    # TODO: add some type checking (e.g. for shape as a tuple of ints) and extract as function
    def _set_shape(self, distribution, shape):
        """
        Sets shape in accordance with distribution.
        """
        self._shape = shape
        if distribution is not None:
            if distribution.shape is not None:
                if shape is None or distribution.shape == shape:
                    self._shape = distribution.shape
                else:
                    raise ValueError(
                        "Shape of distribution and given shape do not match."
                    )
            else:
                self.distribution.reshape(newshape=shape)

    def _set_dtype(self, distribution, dtype):
        """
        Sets dtype in accordance with distribution.
        """
        self._dtype = dtype
        if dtype is not None:
            if isinstance(dtype, np.dtype):
                self._dtype = dtype
            else:
                self._dtype = np.dtype(dtype)
        if distribution is not None:
            if self.dtype is None:
                self._dtype = distribution.dtype
            elif self.dtype != distribution.dtype:
                # Change distribution dtype if random variable type is different
                distribution.dtype = dtype
        if self._dtype is None:
            raise ValueError("No 'dtype' set.")

    def __repr__(self):
        if self.dtype is None:
            dt = "unspecified dtype"
        else:
            dt = "dtype=" + str(self.dtype)
        return "<%s %s with %s>" % (str(self.shape), self.__class__.__name__, dt)

    @property
    def distribution(self):
        """Distribution of random variable."""
        return self._distribution

    def mean(self):
        """Expected value of the random variable."""
        if self._distribution is not None:
            try:
                return self._distribution.mean()
            except KeyError:
                raise NotImplementedError(
                    "Underlying {} has no mean.".format(
                        type(self._distribution).__name__
                    )
                )
        else:
            raise NotImplementedError("No underlying distribution specified.")

    def cov(self):
        """Covariance operator of the random variable"""
        if self._distribution is not None:
            try:
                return self._distribution.cov()
            except KeyError:
                raise NotImplementedError(
                    "Underlying {} has no kernels.".format(
                        type(self._distribution).__name__
                    )
                )
        else:
            raise NotImplementedError("No underlying distribution specified.")

    @property
    def shape(self):
        """Shape of realizations of the random variable."""
        return self._shape

    @property
    def dtype(self):
        """Data type of (elements of) a realization of this random variable."""
        return self._dtype

    @property
    def random_state(self):
        """Random state of the random variable induced by its ``distribution``."""
        return self._distribution.random_state

    @random_state.setter
    def random_state(self, seed):
        """ Get or set the RandomState object of the underlying distribution.

        This can be either None or an existing :class:`~numpy.random.RandomState` object.
        If None (or np.random), use the :class:`~numpy.random.RandomState` singleton used by np.random.
        If already a :class:`~numpy.random.RandomState` instance, use it.
        If an int, use a new :class:`~numpy.random.RandomState` instance seeded with seed.
        """
        self.distribution._random_state = scipy._lib._util.check_random_state(seed)

    def sample(self, size=()):
        """
        Draw realizations from a random variable.

        Parameters
        ----------
        size : tuple
            Size of the drawn sample of realizations.

        Returns
        -------
        sample : array-like
            Sample of realizations with the given ``size`` and the inherent ``shape``.
        """
        if self._distribution is not None:
            return self._distribution.sample(size=size)
        else:
            raise NotImplementedError("No sampling method provided.")

    def reshape(self, newshape):
        """
        Give a new shape to a random variable.

        Parameters
        ----------
        newshape : int or tuple of ints
            New shape for the random variable. It must be compatible with the original shape.

        Returns
        -------
        reshaped_rv : ``self`` with the new dimensions of ``shape``.
        """
        self._shape = newshape
        self._distribution.reshape(newshape=newshape)
        return self

    # Binary arithmetic operations

    # The below prevents numpy from calling elementwise arithmetic
    # operations allowing expressions like: y = np.array([1, 1]) + RV
    # to call the arithmetic operations defined by RandomVariable
    # instead of elementwise. Thus no array of RandomVariables but a
    # RandomVariable with the correct shape is returned.
    __array_ufunc__ = None

    def _rv_from_binary_operation(self, other, op):
        """
        Create a new random variable by applying a binary operation.

        Parameters
        ----------
        other : object
        op : callable
            Binary operation.

        Returns
        -------
        combined_rv : RandomVariable
            Random variable resulting from ``op``.

        """
        other_rv = asrandvar(other)
        combined_rv = RandomVariable(
            distribution=op(self.distribution, other_rv.distribution)
        )
        return combined_rv

    def __add__(self, other):
        return self._rv_from_binary_operation(other=other, op=operator.add)

    def __sub__(self, other):
        return self._rv_from_binary_operation(other=other, op=operator.sub)

    def __mul__(self, other):
        return self._rv_from_binary_operation(other=other, op=operator.mul)

    def __matmul__(self, other):
        return self._rv_from_binary_operation(other=other, op=operator.matmul)

    def __truediv__(self, other):
        return self._rv_from_binary_operation(other=other, op=operator.truediv)

    def __pow__(self, power, modulo=None):
        return self._rv_from_binary_operation(other=power, op=operator.pow)

    # Binary arithmetic operations with reflected (swapped) operands
    def __radd__(self, other):
        other_rv = asrandvar(other)
        return other_rv._rv_from_binary_operation(other=self, op=operator.add)

    def __rsub__(self, other):
        other_rv = asrandvar(other)
        return other_rv._rv_from_binary_operation(other=self, op=operator.sub)

    def __rmul__(self, other):
        other_rv = asrandvar(other)
        return other_rv._rv_from_binary_operation(other=self, op=operator.mul)

    def __rmatmul__(self, other):
        other_rv = asrandvar(other)
        return other_rv._rv_from_binary_operation(other=self, op=operator.matmul)

    def __rtruediv__(self, other):
        other_rv = asrandvar(other)
        return other_rv._rv_from_binary_operation(other=self, op=operator.truediv)

    def __rpow__(self, power, modulo=None):
        other_rv = asrandvar(power)
        return other_rv._rv_from_binary_operation(other=self, op=operator.pow)

    # Augmented arithmetic assignments (+=, -=, *=, ...) attempting to do the operation in place
    # TODO: needs setter functions for properties `shape` and `dtype` to do in place
    def __iadd__(self, other):
        return NotImplemented

    def __isub__(self, other):
        return NotImplemented

    def __imul__(self, other):
        return NotImplemented

    def __imatmul__(self, other):
        return NotImplemented

    def __itruediv__(self, other):
        return NotImplemented

    def __ipow__(self, power, modulo=None):
        return NotImplemented

    # Unary arithmetic operations
    def __neg__(self):
        return RandomVariable(
            shape=self.shape,
            dtype=self.dtype,
            distribution=operator.neg(self.distribution),
        )

    def __pos__(self):
        return RandomVariable(
            shape=self.shape,
            dtype=self.dtype,
            distribution=operator.pos(self.distribution),
        )

    def __abs__(self):
        return RandomVariable(
            shape=self.shape,
            dtype=self.dtype,
            distribution=operator.abs(self.distribution),
        )


def asrandvar(obj):
    """
    Return ``obj`` as a :class:`RandomVariable`.

    Converts scalars, (sparse) arrays or distribution classes to a :class:`RandomVariable`.

    Parameters
    ----------
    obj : object
        Argument to be represented as a :class:`RandomVariable`.

    Returns
    -------
    rv : RandomVariable
        The object `obj` as a :class:`RandomVariable`.

    See Also
    --------
    RandomVariable : Class representing random variables.

    Examples
    --------
    >>> from scipy.stats import bernoulli
    >>> from probnum.prob import asrandvar
    >>> bern = bernoulli(p=0.5)
    >>> bern.random_state = 42  # Seed for reproducibility
    >>> b = asrandvar(bern)
    >>> b.sample(size=5)
    array([1, 1, 1, 0, 0])
    """
    # RandomVariable
    if isinstance(obj, RandomVariable):
        return obj
    # Scalar
    elif np.isscalar(obj):
        return RandomVariable(
            shape=(), dtype=type(obj), distribution=Dirac(support=obj)
        )
    # Numpy array, sparse array or Linear Operator
    elif isinstance(
        obj, (np.ndarray, scipy.sparse.spmatrix, scipy.sparse.linalg.LinearOperator)
    ):
        return RandomVariable(
            shape=obj.shape, dtype=obj.dtype, distribution=Dirac(support=obj)
        )
    # Scipy random variable
    elif isinstance(obj, scipy.stats._distn_infrastructure.rv_frozen) or isinstance(
        obj, scipy.stats._multivariate.multi_rv_frozen
    ):
        return _scipystats_to_rv(scipydist=obj)
    else:
        raise ValueError(
            "Argument of type {} cannot be converted to a random variable.".format(
                type(obj)
            )
        )


def _scipystats_to_rv(scipydist):
    """
    Transform SciPy distributions to Probnum :class:`RandomVariable`s.

    Parameters
    ----------
    scipydist : scipy.stats._distn_infrastructure.rv_frozen or scipy.stats._multivariate.multi_rv_frozen
        SciPy distribution.

    Returns
    -------
    dist : RandomVariable
        ProbNum random variable.

    """
    # Univariate distributions (implemented in this package)
    if isinstance(scipydist, scipy.stats._distn_infrastructure.rv_frozen):
        # Normal distribution
        if scipydist.dist.name == "norm":
            return RandomVariable(
                dtype=float,
                distribution=Normal(
                    mean=scipydist.mean(),
                    cov=scipydist.var(),
                    random_state=scipydist.random_state,
                ),
            )
    # Multivariate distributions (implemented in this package)
    elif isinstance(scipydist, scipy.stats._multivariate.multi_rv_frozen):
        # Multivariate normal
        if scipydist.__class__.__name__ == "multivariate_normal_frozen":
            return RandomVariable(
                shape=scipydist.mean.shape,
                dtype=scipydist.mean.dtype,
                distribution=Normal(
                    mean=scipydist.mean,
                    cov=scipydist.cov,
                    random_state=scipydist.random_state,
                ),
            )
    # Generic distributions
    if hasattr(scipydist, "pmf"):
        pdf = getattr(scipydist, "pmf", None)
        logpdf = getattr(scipydist, "logpmf", None)
    else:
        pdf = getattr(scipydist, "pdf", None)
        logpdf = getattr(scipydist, "logpdf", None)
    rvsample = np.squeeze(scipydist.rvs(1))
    if np.shape(rvsample) == ():
        rvdtype = type(rvsample)
        rvshape = ()
    else:
        rvdtype = rvsample.dtype
        rvshape = rvsample.shape
    return RandomVariable(
        dtype=rvdtype,
        shape=rvshape,
        distribution=Distribution(
            parameters={},
            pdf=pdf,
            logpdf=logpdf,
            cdf=getattr(scipydist, "cdf", None),
            logcdf=getattr(scipydist, "logcdf", None),
            sample=getattr(scipydist, "rvs", None),
            mean=getattr(scipydist, "mean", None),
            cov=getattr(scipydist, "var", None),
            random_state=getattr(scipydist, "random_state", None),
        ),
    )



"""
Probability and statistics.

This package implements functionality related to probability theory and statistics such as random variables and
distributions. Random variables are the primary in- and outputs of probabilistic numerical methods. A generic signature
of such methods looks like this:

.. highlight:: python
.. code-block:: python

    randvar_out, info = probnum_method(problem, randvar_in, **kwargs)

"""

from .randomvariable import *
from .distributions import *

# Public classes and functions. Order is reflected in documentation.
__all__ = ["RandomVariable", "Distribution", "Dirac", "Normal", "asrandvar"]

# Set correct module paths. Corrects links and module paths in documentation.
RandomVariable.__module__ = "probnum.prob"
Distribution.__module__ = "probnum.prob"



"""
Finite dimensional linear operators.

This module defines classes and methods that implement finite dimensional linear operators. It can be used to do linear
algebra with (structured) matrices without explicitly representing them in memory. This often allows for the definition
of a more efficient matrix-vector product. Linear operators can be applied, added, multiplied, transposed, and more as
one would expect from matrix algebra.

Several algorithms in the :mod:`probnum.linalg` library are able to operate on :class:`LinearOperator` instances.
"""
import warnings

import numpy as np
import scipy.sparse.linalg
import scipy.sparse.linalg.interface


class LinearOperator(scipy.sparse.linalg.LinearOperator):
    """
    Finite dimensional linear operators.

    This class provides a way to define finite dimensional linear operators without explicitly constructing a matrix
    representation. Instead it suffices to define a matrix-vector product and a shape attribute. This avoids unnecessary
    memory usage and can often be more convenient to derive.

    LinearOperator instances can be multiplied, added and exponentiated. This happens lazily: the result of these
    operations is a new, composite LinearOperator, that defers linear operations to the original operators and combines
    the results.

    To construct a concrete LinearOperator, either pass appropriate callables to the constructor of this class, or
    subclass it.

    A subclass must implement either one of the methods ``_matvec`` and ``_matmat``, and the
    attributes/properties ``shape`` (pair of integers) and ``dtype`` (may be ``None``). It may call the ``__init__`` on
    this class to have these attributes validated. Implementing ``_matvec`` automatically implements ``_matmat`` (using
    a naive algorithm) and vice-versa.

    Optionally, a subclass may implement ``_rmatvec`` or ``_adjoint`` to implement the Hermitian adjoint (conjugate
    transpose). As with ``_matvec`` and ``_matmat``, implementing either ``_rmatvec`` or ``_adjoint`` implements the
    other automatically. Implementing ``_adjoint`` is preferable; ``_rmatvec`` is mostly there for backwards
    compatibility.

    This class inherits from :class:`scipy.sparse.linalg.LinearOperator`.

    Parameters
    ----------
    shape : tuple
        Matrix dimensions (M, N).
    matvec : callable f(v)
        Returns :math:`A v`.
    rmatvec : callable f(v)
        Returns :math:`A^H v`, where :math:`A^H` is the conjugate transpose of :math:`A`.
    matmat : callable f(V)
        Returns :math:`AV`, where :math:`V` is a dense matrix with dimensions (N, K).
    dtype : dtype
        Data type of the operator.
    rmatmat : callable f(V)
        Returns :math:`A^H V`, where :math:`V` is a dense matrix with dimensions (M, K).

    See Also
    --------
    aslinop : Transform into a LinearOperator.

    Examples
    --------
    >>> import numpy as np
    >>> from probnum.linalg.linops import LinearOperator
    >>> def mv(v):
    ...     return np.array([2 * v[0] - v[1], 3 * v[1]])
    ...
    >>> A = LinearOperator(shape=(2, 2), matvec=mv)
    >>> A
    <2x2 _CustomLinearOperator with dtype=float64>
    >>> A.matvec(np.array([1., 2.]))
    array([0., 6.])
    >>> A @ np.ones(2)
    array([1., 3.])
    """

    def __new__(cls, *args, **kwargs):
        if cls is LinearOperator:
            # _CustomLinearOperator factory
            return super().__new__(_CustomLinearOperator)
        else:
            obj = super().__new__(cls)

            if (
                type(obj)._matvec == scipy.sparse.linalg.LinearOperator._matvec
                and type(obj)._matmat == scipy.sparse.linalg.LinearOperator._matmat
            ):
                warnings.warn(
                    "LinearOperator subclass should implement"
                    " at least one of _matvec and _matmat.",
                    category=RuntimeWarning,
                    stacklevel=2,
                )

            return obj

    # Overload arithmetic operators to give access to newly implemented functions (e.g. todense())
    def __rmul__(self, x):
        if np.isscalar(x):
            return _ScaledLinearOperator(self, x)
        else:
            return NotImplemented

    def __pow__(self, p):
        if np.isscalar(p):
            return _PowerLinearOperator(self, p)
        else:
            return NotImplemented

    def __add__(self, x):
        if isinstance(x, LinearOperator):
            return _SumLinearOperator(self, x)
        else:
            return NotImplemented

    def __neg__(self):
        return _ScaledLinearOperator(self, -1)

    # The below methods are overloaded to allow dot products with random variables
    def dot(self, x):
        """Matrix-matrix or matrix-vector multiplication.
        Parameters
        ----------
        x : array_like
            1-d or 2-d array, representing a vector or matrix.
        Returns
        -------
        Ax : array
            1-d or 2-d array (depending on the shape of x) that represents
            the result of applying this linear operator on x.
        """
        if isinstance(x, LinearOperator):
            return _ProductLinearOperator(self, x)
        elif np.isscalar(x):
            return _ScaledLinearOperator(self, x)
        else:
            if len(x.shape) == 1 or len(x.shape) == 2 and x.shape[1] == 1:
                return self.matvec(x)
            elif len(x.shape) == 2:
                return self.matmat(x)
            else:
                raise ValueError(
                    "Expected 1-d or 2-d array, matrix or random variable, got %r." % x
                )

    def matvec(self, x):
        """Matrix-vector multiplication.
        Performs the operation y=A*x where A is an MxN linear
        operator and x is a 1-d array or random variable.

        Parameters
        ----------
        x : {matrix, ndarray, RandomVariable}
            An array or RandomVariable with shape (N,) or (N,1).
        Returns
        -------
        y : {matrix, ndarray}
            A matrix or ndarray or RandomVariable with shape (M,) or (M,1) depending
            on the type and shape of the x argument.
        Notes
        -----
        This matvec wraps the user-specified matvec routine or overridden
        _matvec method to ensure that y has the correct shape and type.
        """
        M, N = self.shape

        if x.shape != (N,) and x.shape != (N, 1):
            raise ValueError("Dimension mismatch.")

        y = self._matvec(x)

        if isinstance(x, np.matrix):
            y = scipy.sparse.sputils.asmatrix(y)

        if isinstance(x, (np.matrix, np.ndarray)):
            if x.ndim == 1:
                y = y.reshape(M)
            elif x.ndim == 2:
                y = y.reshape(M, 1)
            else:
                raise ValueError("Invalid shape returned by user-defined matvec().")
        # TODO: can be shortened once RandomVariable implements a reshape method
        elif y.shape[0] != M:
            raise ValueError("Invalid shape returned by user-defined matvec().")

        return y

    def transpose(self):
        """
        Transpose this linear operator.

        Can be abbreviated self.T instead of self.transpose().
        """
        return self._transpose()

    T = property(transpose)

    def _transpose(self):
        """ Default implementation of _transpose; defers to rmatvec + conj"""
        return _TransposedLinearOperator(self)

    def todense(self):
        """
        Dense matrix representation of the linear operator.

        This method can be computationally very costly depending on the shape of the linear operator. Use with caution.

        Returns
        -------
        matrix : np.ndarray
            Matrix representation of the linear operator.
        """
        return self.matmat(np.eye(self.shape[1], dtype=self.dtype))

    def inv(self):
        """Inverse of the linear operator."""
        raise NotImplementedError

    # TODO: implement operations (eigs, cond, det, logabsdet, trace, ...)
    def rank(self):
        """Rank of the linear operator."""
        raise NotImplementedError

    def eigvals(self):
        """Eigenvalue spectrum of the linear operator."""
        raise NotImplementedError

    def cond(self, p=None):
        """
        Compute the condition number of the linear operator.

        The condition number of the linear operator with respect to the ``p`` norm. It measures how much the solution
        :math:`x` of the linear system :math:`Ax=b` changes with respect to small changes in :math:`b`.

        Parameters
        ----------
        p : {None, 1, , 2, , inf, 'fro'}, optional
            Order of the norm:

            =======  ============================
            p        norm for matrices
            =======  ============================
            None     2-norm, computed directly via singular value decomposition
            'fro'    Frobenius norm
            np.inf   max(sum(abs(x), axis=1))
            1        max(sum(abs(x), axis=0))
            2        2-norm (largest sing. value)
            =======  ============================

        Returns
        -------
        cond : {float, inf}
            The condition number of the linear operator. May be infinite.
        """
        raise NotImplementedError

    def det(self):
        """Determinant of the linear operator."""
        raise NotImplementedError

    def logabsdet(self):
        """Log absolute determinant of the linear operator."""
        raise NotImplementedError

    def trace(self):
        """
        Trace of the linear operator.

        Computes the trace of a square linear operator :math:`\\text{tr}(A) = \\sum_{i-1}^n A_ii`.

        Returns
        -------
        trace : float
            Trace of the linear operator.

        Raises
        ------
        ValueError : If :meth:`trace` is called on a non-square matrix.
        """
        if self.shape[0] != self.shape[1]:
            raise ValueError("The trace is only defined for square linear operators.")
        else:
            _identity = np.eye(self.shape[0])
            trace = 0.0
            for i in range(self.shape[0]):
                trace += np.squeeze(
                    _identity[np.newaxis, i, :]
                    @ self.matvec(_identity[i, :, np.newaxis])
                )
            return trace


class _CustomLinearOperator(
    scipy.sparse.linalg.interface._CustomLinearOperator, LinearOperator
):
    """Linear operator defined in terms of user-specified operations."""

    def __init__(
        self, shape, matvec, rmatvec=None, matmat=None, rmatmat=None, dtype=None
    ):
        super().__init__(
            shape=shape,
            matvec=matvec,
            rmatvec=rmatvec,
            matmat=matmat,
            rmatmat=rmatmat,
            dtype=dtype,
        )


# TODO: inheritance from _TransposedLinearOperator causes dependency on scipy>=1.4, maybe implement our own instead?
class _TransposedLinearOperator(
    scipy.sparse.linalg.interface._TransposedLinearOperator, LinearOperator
):
    """Transposition of a linear operator."""

    def __init__(self, A):
        self.A = A
        super().__init__(A=A)

    def todense(self):
        return self.A.todense().T

    def inv(self):
        return self.A.inv().T


class _SumLinearOperator(
    scipy.sparse.linalg.interface._SumLinearOperator, LinearOperator
):
    """Sum of two linear operators."""

    def __init__(self, A, B):
        self.A = A
        self.B = B
        super().__init__(A=A, B=B)

    def todense(self):
        return self.A.todense() + self.B.todense()

    def inv(self):
        return self.A.inv() + self.B.inv()

    def trace(self):
        return self.A.trace() + self.B.trace()


class _ProductLinearOperator(
    scipy.sparse.linalg.interface._ProductLinearOperator, LinearOperator
):
    """(Operator) Product of two linear operators."""

    def __init__(self, A, B):
        self.A = A
        self.B = B
        super().__init__(A=A, B=B)

    def todense(self):
        return self.A.todense() @ self.B.todense()


class _ScaledLinearOperator(
    scipy.sparse.linalg.interface._ScaledLinearOperator, LinearOperator
):
    """Linear operator scaled with a scalar."""

    def __init__(self, A, alpha):
        super().__init__(A=A, alpha=alpha)

    def todense(self):
        A, alpha = self.args
        return alpha * A.todense()

    def inv(self):
        A, alpha = self.args
        return _ScaledLinearOperator(A.inv(), 1 / alpha)

    def trace(self):
        A, alpha = self.args
        return alpha * A.trace()


class _PowerLinearOperator(
    scipy.sparse.linalg.interface._PowerLinearOperator, LinearOperator
):
    """Linear operator raised to a non-negative integer power."""

    def __init__(self, A, p):
        super().__init__(A=A, p=p)


class Diagonal(LinearOperator):
    """
    A linear operator representing the diagonal from another linear operator.

    Parameters
    ----------
    Op : LinearOperator
        Linear operator of which to represent the diagonal.
    """

    # TODO: should this be an operator itself or a function of a LinearOperator?
    #   - a function allows subclasses (e.g. MatrixMult) to implement more efficient versions than n products e_i A e_i
    def __init__(self, Op):
        raise NotImplementedError


class ScalarMult(LinearOperator):
    """
    A linear operator representing scalar multiplication.

    Parameters
    ----------
    shape : tuple
        Matrix dimensions (M, N).
    scalar : float
        Scalar to multiply by.
    """

    def __init__(self, shape, scalar):
        self.scalar = scalar
        super().__init__(shape=shape, dtype=float)

    def _matvec(self, x):
        return self.scalar * x

    def _matmat(self, X):
        return self.scalar * X

    def todense(self):
        return np.eye(self.shape[0]) * self.scalar

    def inv(self):
        return ScalarMult(shape=self.shape, scalar=1 / self.scalar)

    # Properties
    def rank(self):
        return np.minimum(self.shape[0], self.shape[1])

    def eigvals(self):
        return np.ones(self.shape[0]) * self.scalar

    def cond(self, p=None):
        return 1

    def det(self):
        return self.scalar ** self.shape[0]

    def logabsdet(self):
        return np.log(np.abs(self.scalar))

    def trace(self):
        return self.scalar * self.shape[0]


class Identity(ScalarMult):
    """
    The identity operator.

    Parameters
    ----------
    shape : int or tuple
        Shape of the identity operator.
    """

    def __init__(self, shape):
        # Check shape
        if np.isscalar(shape):
            _shape = (shape, shape)
        elif shape[0] != shape[1]:
            raise ValueError("The identity operator must be square.")
        else:
            _shape = shape
        # Initiator of super class
        super().__init__(shape=_shape, scalar=1.0)

    def todense(self):
        return np.eye(self.shape[0])

    def inv(self):
        return self

    # Properties
    def rank(self):
        return self.shape[0]

    def eigvals(self):
        return np.ones(self.shape[0])

    def cond(self, p=None):
        return 1

    def det(self):
        return 1.0

    def logabsdet(self):
        return 0.0

    def trace(self):
        return self.shape[0]


class MatrixMult(scipy.sparse.linalg.interface.MatrixLinearOperator, LinearOperator):
    """
    A linear operator defined via a matrix.

    Parameters
    ----------
    A : array-like or scipy.sparse.spmatrix
        The explicit matrix.
    """

    def __init__(self, A):
        super().__init__(A=A)

    def _matvec(self, x):
        return self.A @ x  # Needed to call __matmul__ instead of np.dot or np.matmul

    def _matmat(self, X):
        return self.A @ X

    def todense(self):
        if isinstance(self.A, scipy.sparse.spmatrix):
            return self.A.todense()
        else:
            return np.asarray(self.A)

    def inv(self):
        if isinstance(self.A, scipy.sparse.spmatrix):
            invmat = scipy.sparse.linalg.inv(self.A)
        else:
            invmat = np.linalg.inv(self.A)
        return MatrixMult(invmat)

    # Arithmetic operations
    # TODO: perform arithmetic operations between MatrixMult operators explicitly

    # Properties
    def rank(self):
        return np.linalg.matrix_rank(self.A)

    def eigvals(self):
        return np.linalg.eigvals(self.A)

    def cond(self, p=None):
        return np.linalg.cond(self.A, p=p)

    def det(self):
        return np.linalg.det(self.A)

    def logabsdet(self):
        sign, logdet = np.linalg.slogdet(self.A)
        return logdet

    def trace(self):
        if self.shape[0] != self.shape[1]:
            raise ValueError("The trace is only defined for square linear operators.")
        else:
            return np.trace(self.A)


def aslinop(A):
    """
    Return `A` as a :class:`LinearOperator`.

    Parameters
    ----------
    A : array-like or LinearOperator or RandomVariable or object
        Argument to be represented as a linear operator. When `A` is an object it needs to have the attributes `.shape`
        and `.matvec`.

    Notes
    -----
    If `A` has no `.dtype` attribute, the data type is determined by calling
    :func:`LinearOperator.matvec()` - set the `.dtype` attribute to prevent this
    call upon the linear operator creation.

    See Also
    --------
    LinearOperator : Class representing linear operators.

    Examples
    --------
    >>> from probnum.linalg.linops import aslinop
    >>> M = np.array([[1,2,3],[4,5,6]], dtype=np.int32)
    >>> aslinop(M)
    <2x3 MatrixMult with dtype=int32>
    """
    if isinstance(A, scipy.sparse.linalg.LinearOperator):
        return A
    elif isinstance(A, (np.ndarray, scipy.sparse.spmatrix)):
        return MatrixMult(A=A)
    else:
        op = scipy.sparse.linalg.aslinearoperator(A)
        return LinearOperator(op)



"""
General Gaussian filters based on approximating intractable
quantities with numerical quadrature.
Examples include the unscented Kalman filter / RTS smoother
which is based on a third degree fully symmetric rule.
"""

import numpy as np

from probnum.filtsmooth.gaussfiltsmooth.gaussfiltsmooth import *
from probnum.prob import RandomVariable, Normal
from probnum.filtsmooth.statespace import *
from probnum.filtsmooth.gaussfiltsmooth.unscentedtransform import *


class UnscentedKalman(GaussFiltSmooth):
    """
    Factory method for Unscented Kalman filters.
    """

    def __new__(cls, dynamod, measmod, initrv, alpha, beta, kappa, **kwargs):
        """ """
        if cls is UnscentedKalman:
            if _cont_disc(dynamod, measmod):
                return _ContDiscUnscentedKalman(
                    dynamod, measmod, initrv, alpha, beta, kappa, **kwargs
                )
            if _disc_disc(dynamod, measmod):
                return _DiscDiscUnscentedKalman(
                    dynamod, measmod, initrv, alpha, beta, kappa, **kwargs
                )
            else:
                errmsg = (
                    "Cannot instantiate Unscented Kalman filter with "
                    "given dynamic model and measurement model."
                )
                raise ValueError(errmsg)
        else:
            return super().__new__(cls)


def _cont_disc(dynamod, measmod):
    """ """
    dyna_is_cont = issubclass(type(dynamod), ContinuousModel)
    meas_is_disc = issubclass(type(measmod), DiscreteModel)
    return dyna_is_cont and meas_is_disc


def _disc_disc(dynamod, measmod):
    """ """
    dyna_is_disc = issubclass(type(dynamod), DiscreteModel)
    meas_is_disc = issubclass(type(measmod), DiscreteModel)
    return dyna_is_disc and meas_is_disc


class _ContDiscUnscentedKalman(UnscentedKalman):
    def __init__(self, dynamod, measmod, initrv, alpha, beta, kappa, **kwargs):
        if not issubclass(type(dynamod), LinearSDEModel):
            raise ValueError(
                "This implementation of "
                "_ContDiscUnscentedKalman "
                "requires a linear dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianModel):
            raise ValueError(
                "_DiscDiscUnscentedKalman requires " "a Gaussian measurement model."
            )
        if "cke_nsteps" in kwargs.keys():
            self.cke_nsteps = kwargs["cke_nsteps"]
        else:
            self.cke_nsteps = 1
        super().__init__(dynamod, measmod, initrv)
        self.ut = UnscentedTransform(self.dynamod.ndim, alpha, beta, kappa)

    def predict(self, start, stop, randvar, **kwargs):
        step = (stop - start) / self.cke_nsteps
        return self.dynamicmodel.chapmankolmogorov(start, stop, step, randvar, **kwargs)

    def update(self, time, randvar, data, **kwargs):
        return _discrete_unskalman_update(
            time, randvar, data, self.measmod, self.ut, **kwargs
        )


class _DiscDiscUnscentedKalman(UnscentedKalman):
    def __init__(self, dynamod, measmod, initrv, alpha, beta, kappa, **kwargs):
        """
        Checks that dynamod and measmod are linear and moves on.
        """
        if not issubclass(type(dynamod), DiscreteGaussianModel):
            raise ValueError(
                "_DiscDiscUnscentedKalman requires " "a Gaussian dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianModel):
            raise ValueError(
                "_DiscDiscUnscentedKalman requires " "a Gaussian measurement model."
            )
        super().__init__(dynamod, measmod, initrv)
        self.ut = UnscentedTransform(self.dynamod.ndim, alpha, beta, kappa)

    def predict(self, start, stop, randvar, **kwargs):
        if issubclass(type(self.dynamod), DiscreteGaussianLinearModel):
            return self._predict_linear(start, randvar, **kwargs)
        else:
            return self._predict_nonlinear(start, randvar, **kwargs)

    def _predict_linear(self, start, randvar, **kwargs):
        """
        Basic Kalman update because model is linear.
        """
        mean, covar = randvar.mean(), randvar.cov()
        if np.isscalar(mean) and np.isscalar(covar):
            mean, covar = mean * np.ones(1), covar * np.eye(1)
        dynamat = self.dynamod.dynamicsmatrix(start, **kwargs)
        forcevec = self.dynamod.force(start, **kwargs)
        diffmat = self.dynamod.diffusionmatrix(start, **kwargs)
        mpred = dynamat @ mean + forcevec
        crosscov = covar @ dynamat.T
        cpred = dynamat @ crosscov + diffmat
        return RandomVariable(distribution=Normal(mpred, cpred)), crosscov

    def _predict_nonlinear(self, start, randvar, **kwargs):
        """
        Executes unscented transform!
        """
        mean, covar = randvar.mean(), randvar.cov()
        if np.isscalar(mean) and np.isscalar(covar):
            mean, covar = mean * np.ones(1), covar * np.eye(1)
        sigmapts = self.ut.sigma_points(mean, covar)
        proppts = self.ut.propagate(start, sigmapts, self.dynamod.dynamics)
        diffmat = self.dynamod.diffusionmatrix(start, **kwargs)
        mpred, cpred, crosscov = self.ut.estimate_statistics(
            proppts, sigmapts, diffmat, mean
        )
        return RandomVariable(distribution=Normal(mpred, cpred)), crosscov

    def update(self, time, randvar, data, **kwargs):
        """ """
        return _discrete_unskalman_update(
            time, randvar, data, self.measmod, self.ut, **kwargs
        )


def _discrete_unskalman_update(time, randvar, data, measmod, ut, **kwargs):
    """
    """
    if issubclass(type(measmod), DiscreteGaussianLinearModel):
        return _update_discrete_linear(time, randvar, data, measmod, **kwargs)
    else:
        return _update_discrete_nonlinear(time, randvar, data, measmod, ut, **kwargs)


def _update_discrete_linear(time, randvar, data, measmod, **kwargs):
    """
    """
    mpred, cpred = randvar.mean(), randvar.cov()
    if np.isscalar(mpred) and np.isscalar(cpred):
        mpred, cpred = mpred * np.ones(1), cpred * np.eye(1)
    measmat = measmod.dynamicsmatrix(time, **kwargs)
    meascov = measmod.diffusionmatrix(time, **kwargs)
    meanest = measmat @ mpred
    covest = measmat @ cpred @ measmat.T + meascov
    ccest = cpred @ measmat.T
    mean = mpred + ccest @ np.linalg.solve(covest, data - meanest)
    cov = cpred - ccest @ np.linalg.solve(covest.T, ccest.T)
    return RandomVariable(distribution=Normal(mean, cov)), covest, ccest, meanest


def _update_discrete_nonlinear(time, randvar, data, measmod, ut, **kwargs):
    """
    """
    mpred, cpred = randvar.mean(), randvar.cov()
    if np.isscalar(mpred) and np.isscalar(cpred):
        mpred, cpred = mpred * np.ones(1), cpred * np.eye(1)
    sigmapts = ut.sigma_points(mpred, cpred)
    proppts = ut.propagate(time, sigmapts, measmod.dynamics)
    meascov = measmod.diffusionmatrix(time, **kwargs)
    meanest, covest, ccest = ut.estimate_statistics(proppts, sigmapts, meascov, mpred)
    mean = mpred + ccest @ np.linalg.solve(covest, data - meanest)
    cov = cpred - ccest @ np.linalg.solve(covest.T, ccest.T)
    return RandomVariable(distribution=Normal(mean, cov)), covest, ccest, meanest



"""
Continuous Markov models implicitly defined through SDEs,
dx(t) = f(t, x(t)) dt + l(t, x(t)) dB(t).
Possible input for StateSpace.dynmod or StateSpace.measmod.

Note
----
ContinuousModel only implements
StateSpaceComponent.sample(). It passes on the
responsibility of implementing chapmankolmogorov() to its
subclasses and creates further abstract methods drift(),
dispersion() as well as a hook for jacobian().

When implementing continuous models we think of
Gauss-Markov models (see linearsde.py)
but do our very best to not enforce Gaussianity.
In the future, this might be more general than Gauss-
Markov models.
"""

from abc import ABC, abstractmethod

import numpy as np


__all__ = ["ContinuousModel"]


class ContinuousModel(ABC):
    """
    Interface for time-continuous
    Markov models of the form
    dx = f(t, x) dt + l(t, x) dBt.

    In the language of dynamic models:
    x(t) : state process
    f(t, x(t)) : drift function
    l(t, x(t)) : dispersion matrix.
    B(t) : Brownian motion with const. diffusion matrix Q.
    """

    def sample(self, start, stop, step, initstate, **kwargs):
        """
        Samples from ``initstate`` at ``start`` to ``stop`` with
        stepsize ``step``.

        Start, stop and step lead to a np.arange-like
        interface.
        Returns a single element at the end of the time, not the
        entire array!
        """
        if type(initstate) != np.ndarray:
            raise ValueError("Init state is not array!")
        times = np.arange(start, stop, step)
        currstate = initstate
        for idx in range(1, len(times)):
            bmsamp = np.random.multivariate_normal(
                np.zeros(len(self.diffusionmatrix)), self.diffusionmatrix * step
            )
            driftvl = self.drift(times[idx - 1], currstate, **kwargs)
            dispvl = self.dispersion(times[idx - 1], currstate, **kwargs)
            currstate = currstate + step * driftvl + dispvl @ bmsamp
        return currstate

    @abstractmethod
    def drift(self, time, state, **kwargs):
        """
        Evaluates f = f(t, x(t)).
        """
        raise NotImplementedError

    @abstractmethod
    def dispersion(self, time, state, **kwargs):
        """
        Evaluates l = l(t, x(t)).
        """
        raise NotImplementedError

    def jacobian(self, time, state, **kwargs):
        """
        Jacobian of drift w.r.t. state: d_x f(t, x(t))
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def diffusionmatrix(self):
        """
        Evaluates Q.
        In 1D, this is \\sigma^2.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def ndim(self):
        """
        Spatial dimension (utility attribute).
        """
        raise NotImplementedError

    def chapmankolmogorov(self, start, stop, step, randvar, **kwargs):
        """
        If available, this returns the closed form solution to the
        Chapman-Kolmogorov equations (CKEs).

        Solutions to the CKEs are important in filtering.

        Available for instance for linear SDEs.
        """
        raise NotImplementedError("Chap.-Kolg. not implemented.")



"""
Gaussian filtering.
"""
import numpy as np

from probnum.prob import RandomVariable, Normal
from probnum.filtsmooth.bayesfiltsmooth import *


class GaussFiltSmooth(BayesFiltSmooth, ABC):
    """
    Interface for Gaussian filtering and smoothing.
    """

    def __init__(self, dynamod, measmod, initrv):
        """        Check that the initial distribution is Gaussian.        """
        if not issubclass(type(initrv.distribution), Normal):
            raise ValueError(
                "Gaussian filters/smoothers need initial "
                "random variables with Normal distribution."
            )
        super().__init__(dynamod, measmod, initrv)

    def filtsmooth(self, dataset, times, **kwargs):
        """
        Apply Gaussian filtering and smoothing to a data set.

        Parameters
        ----------
        dataset : array_like, shape (N, M)
            Data set that is filtered.
        times : array_like, shape (N,)
            Temporal locations of the data points.
        kwargs : ???

        Returns
        -------
        ndarray, shape (N, M)
            Means of the smoothed output
        ndarray, shape (N, M, M)
            Covariances of the smoothed output.
        """
        dataset, times = np.asarray(dataset), np.asarray(times)
        filtered_means, filtered_covs = self.filter(dataset, times, **kwargs)
        smoothed_means, smoothed_covs = self.smooth(
            filtered_means, filtered_covs, times, **kwargs
        )
        return smoothed_means, smoothed_covs

    def filter(self, dataset, times, **kwargs):
        """
        Apply Gaussian filtering (no smoothing!) to a data set.

        Parameters
        ----------
        dataset : array_like, shape (N, M)
            Data set that is filtered.
        times : array_like, shape (N,)
            Temporal locations of the data points.
        kwargs : ???

        Returns
        -------
        ndarray, shape (N, M)
            Means of the filtered output
        ndarray, shape (N, M, M)
            Covariances of the filtered output.
        """
        dataset, times = np.asarray(dataset), np.asarray(times)
        filtrv = self.initialrandomvariable
        means = [filtrv.mean()]
        covs = [filtrv.cov()]
        for idx in range(1, len(times)):
            filtrv = self.filter_step(
                start=times[idx - 1],
                stop=times[idx],
                randvar=filtrv,
                data=dataset[idx - 1],
                **kwargs
            )
            means.append(filtrv.mean())
            covs.append(filtrv.cov())
        return np.array(means), np.array(covs)

    def smooth(self, filtmeans, filtcovs, times, **kwargs):
        """
        Apply Gaussian smoothing to a set of filtered means and covariances.

        Parameters
        ----------
        filtmeans : array_like, shape (N, M)
            Means of the filter solution.
        filtcovs : array_like, shape (N, M, M)
            Covariances of the filter solution.
        times : array_like, shape (N,)
            Temporal locations of the filter solution (i.e. locations of
            the data points---the data points are not needed anymore)
        kwargs : ???

        Returns
        -------
        ndarray, shape (N, M)
            Means of the smoothed filter solution
        ndarray, shape (N, M, M)
            Covariances of the smoothed filter solution

        """
        means, covs = np.zeros(filtmeans.shape), np.zeros(filtcovs.shape)
        means[-1], covs[-1] = filtmeans[-1], filtcovs[-1]
        smoothed_rv = RandomVariable(distribution=Normal(filtmeans[-1], filtcovs[-1]))
        for idx in reversed(range(1, len(times))):
            unsmoothed_rv = RandomVariable(
                distribution=Normal(filtmeans[idx - 1], filtcovs[idx - 1])
            )
            pred_rv, ccov = self.predict(
                times[idx - 1], times[idx], unsmoothed_rv, **kwargs
            )
            smoothed_rv = self.smooth_step(unsmoothed_rv, pred_rv, smoothed_rv, ccov)
            means[idx - 1], covs[idx - 1] = smoothed_rv.mean(), smoothed_rv.cov()
        return means, covs

    def filter_step(self, start, stop, randvar, data, **kwargs):
        """
        A single filter step.

        Consists of a prediction step (t -> t+1) and an update step (at t+1).

        Parameters
        ----------
        start : float
            Predict FROM this time point.
        stop : float
            Predict TO this time point.
        randvar : RandomVariable
            Predict based on this random variable. For instance, this can be the result of a previous call to filter_step.
        data : array_like
            Compute the update based on this data.

        Returns
        -------
        RandomVariable
            Resulting filter estimate after the single step.
        """
        data = np.asarray(data)
        predrv, _ = self.predict(start, stop, randvar, **kwargs)
        filtrv, _, _, _ = self.update(stop, predrv, data)
        return filtrv

    def smooth_step(self, unsmoothed_rv, pred_rv, smoothed_rv, crosscov):
        """
        A single smoother step.

        Consists of predicting from the filtering distribution at time t
        to time t+1 and then updating based on the discrepancy to the
        smoothing solution at time t+1.

        Parameters
        ----------
        unsmoothed_rv : RandomVariable
            Filtering distribution at time t.
        pred_rv : RandomVariable
            Prediction at time t+1 of the filtering distribution at time t.
        smoothed_rv : RandomVariable
            Smoothing distribution at time t+1.
        crosscov : array_like
            Cross-covariance between unsmoothed_rv and pred_rv as
            returned by predict().
        """
        crosscov = np.asarray(crosscov)
        initmean, initcov = unsmoothed_rv.mean(), unsmoothed_rv.cov()
        predmean, predcov = pred_rv.mean(), pred_rv.cov()
        currmean, currcov = smoothed_rv.mean(), smoothed_rv.cov()
        if np.isscalar(predmean) and np.isscalar(predcov):
            predmean = predmean * np.ones(1)
            predcov = predcov * np.eye(1)
        res = currmean - predmean
        newmean = initmean + crosscov @ np.linalg.solve(predcov, res)
        firstsolve = crosscov @ np.linalg.solve(predcov, currcov - predcov)
        secondsolve = crosscov @ np.linalg.solve(predcov, firstsolve.T)
        newcov = initcov + secondsolve.T
        return RandomVariable(distribution=Normal(newmean, newcov))

    @abstractmethod
    def predict(self, start, stop, randvar, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def update(self, time, randvar, data, **kwargs):
        raise NotImplementedError



"""
Kalman filtering and (Rauch-Tung-Striebel) smoothing for
continuous-discrete and discrete-discrete state space models.
"""

import numpy as np
from probnum.filtsmooth.gaussfiltsmooth.gaussfiltsmooth import *
from probnum.prob import RandomVariable, Normal
from probnum.filtsmooth.statespace import *


class Kalman(GaussFiltSmooth):
    """
    Kalman filtering and smoothing for continuous-discrete and
    discrete-discrete state space models.
    """

    def __new__(cls, dynamod, measmod, initrv, **kwargs):
        """
        Factory method for Kalman filtering and smoothing.

        Depending on whether the dynamic model is continuous or
        discrete, either a continuous-discrete Kalman object or a
        discrete-discrete Kalman object is created.
        """
        if cls is Kalman:
            if _cont_disc(dynamod, measmod):
                return _ContDiscKalman(dynamod, measmod, initrv, **kwargs)
            if _disc_disc(dynamod, measmod):
                return _DiscDiscKalman(dynamod, measmod, initrv)
            else:
                errmsg = (
                    "Cannot instantiate Kalman object with given "
                    "dynamic model and measurement model."
                )
                raise ValueError(errmsg)
        else:
            return super().__new__(cls)


def _cont_disc(dynamod, measmod):
    """Checks whether the state space model is continuous-discrete."""
    dyna_is_cont = issubclass(type(dynamod), ContinuousModel)
    meas_is_disc = issubclass(type(measmod), DiscreteModel)
    return dyna_is_cont and meas_is_disc


def _disc_disc(dynamod, measmod):
    """Checks whether the state space model is discrete-discrete."""
    dyna_is_disc = issubclass(type(dynamod), DiscreteModel)
    meas_is_disc = issubclass(type(measmod), DiscreteModel)
    return dyna_is_disc and meas_is_disc


class _ContDiscKalman(Kalman):
    """
    Provides predict() and update() methods for Kalman filtering and
    smoothing on continuous-discrete state space models.
    """

    def __init__(self, dynamod, measmod, initrv, **kwargs):
        """
        Checks that dynamod and measmod are linear and moves on.
        """
        if not issubclass(type(dynamod), LinearSDEModel):
            raise ValueError(
                "ContinuosDiscreteKalman requires " "a linear dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianLinearModel):
            raise ValueError(
                "DiscreteDiscreteKalman requires " "a linear measurement model."
            )
        if "cke_nsteps" in kwargs.keys():
            self.cke_nsteps = kwargs["cke_nsteps"]
        else:
            self.cke_nsteps = 1
        super().__init__(dynamod, measmod, initrv)

    def predict(self, start, stop, randvar, **kwargs):
        """ """
        step = (stop - start) / self.cke_nsteps
        return self.dynamicmodel.chapmankolmogorov(start, stop, step, randvar, **kwargs)

    def update(self, time, randvar, data, **kwargs):
        """ """
        return _discrete_kalman_update(
            time, randvar, data, self.measurementmodel, **kwargs
        )


class _DiscDiscKalman(Kalman):
    """
    Provides predict() and update() methods for Kalman filtering and
    smoothing on discrete-discrete state space models.
    """

    def __init__(self, dynamod, measmod, initrv):
        """Checks that dynamod and measmod are linear and moves on."""
        if not issubclass(type(dynamod), DiscreteGaussianLinearModel):
            raise ValueError(
                "ContinuousDiscreteKalman requires " "a linear dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianLinearModel):
            raise ValueError(
                "DiscreteDiscreteKalman requires " "a linear measurement model."
            )
        super().__init__(dynamod, measmod, initrv)

    def predict(self, start, stop, randvar, **kwargs):
        """Prediction step for discrete-discrete Kalman filtering."""
        mean, covar = randvar.mean(), randvar.cov()
        if np.isscalar(mean) and np.isscalar(covar):
            mean, covar = mean * np.ones(1), covar * np.eye(1)
        dynamat = self.dynamicmodel.dynamicsmatrix(start, **kwargs)
        forcevec = self.dynamicmodel.force(start, **kwargs)
        diffmat = self.dynamicmodel.diffusionmatrix(start, **kwargs)
        mpred = dynamat @ mean + forcevec
        ccpred = covar @ dynamat.T
        cpred = dynamat @ ccpred + diffmat
        return RandomVariable(distribution=Normal(mpred, cpred)), ccpred

    def update(self, time, randvar, data, **kwargs):
        """Update step of discrete Kalman filtering"""
        return _discrete_kalman_update(
            time, randvar, data, self.measurementmodel, **kwargs
        )


def _discrete_kalman_update(time, randvar, data, measurementmodel, **kwargs):
    """Discrete Kalman update."""
    mpred, cpred = randvar.mean(), randvar.cov()
    if np.isscalar(mpred) and np.isscalar(cpred):
        mpred, cpred = mpred * np.ones(1), cpred * np.eye(1)
    measmat = measurementmodel.dynamicsmatrix(time, **kwargs)
    meascov = measurementmodel.diffusionmatrix(time, **kwargs)
    meanest = measmat @ mpred
    covest = measmat @ cpred @ measmat.T + meascov
    ccest = cpred @ measmat.T
    mean = mpred + ccest @ np.linalg.solve(covest, data - meanest)
    cov = cpred - ccest @ np.linalg.solve(covest.T, ccest.T)
    return (RandomVariable(distribution=Normal(mean, cov)), covest, ccest, meanest)



"""Utility functions for arrays and the like."""

import numpy as np
import scipy.sparse
import probnum.prob

__all__ = [
    "atleast_1d",
    "atleast_2d",
    "as_colvec",
    "assert_is_1d_ndarray",
    "assert_is_2d_ndarray",
]


def atleast_1d(*rvs):
    """
    Convert arrays or random variables to arrays or random variables
    with at least one dimension.

    Scalar inputs are converted to 1-dimensional arrays, whilst
    higher-dimensional inputs are preserved. Sparse arrays are not
    transformed.

    Parameters
    ----------
    rvs: array-like or RandomVariable
        One or more input random variables or arrays.

    Returns
    -------
    res : array-like or list
        An array / random variable or list of arrays / random variables,
        each with ``a.ndim >= 1``.

    See Also
    --------
    atleast_2d

    """
    res = []
    for rv in rvs:
        if isinstance(rv, scipy.sparse.spmatrix):
            result = rv
        elif isinstance(rv, np.ndarray):
            result = np.atleast_1d(rv)
        elif isinstance(rv, probnum.prob.RandomVariable):
            raise NotImplementedError
        else:
            result = rv
        res.append(result)
    if len(res) == 1:
        return res[0]
    else:
        return res


def atleast_2d(*rvs):
    """
    View inputs as arrays with at least two dimensions.

    Parameters
    ----------
    arys1, arys2, ... : array-like or RandomVariable
        One or more random variables or array-like sequences. Non-array
        inputs are converted to arrays or random variables. Arrays or
        random variables that already have two or more dimensions are
        preserved.

    Returns
    -------
    rvs: array-like or list
        An array, random variable or a list of arrays / random
        variables, each with ``a.ndim >= 2``.

    See Also
    --------
    atleast_1d
    """
    res = []
    for rv in rvs:
        if isinstance(rv, scipy.sparse.spmatrix):
            result = rv
        elif isinstance(rv, np.ndarray):
            result = np.atleast_2d(rv)
        elif isinstance(rv, probnum.prob.RandomVariable):
            raise NotImplementedError
        else:
            result = rv
        res.append(result)
    if len(res) == 1:
        return res[0]
    else:
        return res


def as_colvec(vec):
    """
    Transform the given vector or random variable to column format.

    Given a vector (or random variable) of dimension (n,) return an array with dimensions
    (n, 1) instead. Higher-dimensional arrays are not changed.

    Parameters
    ----------
    vec : np.ndarray or RandomVariable
        Vector, array or random variable to be viewed as a column vector.

    Returns
    -------
    vec2d : np.ndarray or RandomVariable
    """
    if isinstance(vec, probnum.prob.RandomVariable):
        if vec.shape != (vec.shape[0], 1):
            vec.reshape(newshape=(vec.shape[0], 1))
    else:
        if vec.ndim == 1:
            return vec[:, None]
    return vec


def assert_is_1d_ndarray(arr):
    """
    Checks whether arr is an (d,) np.ndarray.

    Used in classic optimization and mcmc, for instance.
    """
    if not isinstance(arr, np.ndarray):
        raise ValueError("Please enter arr of shape (d,)")
    elif len(arr.shape) != 1:
        raise ValueError("Please enter arr of shape (d,)")


def assert_is_2d_ndarray(arr):
    """
    Checks whether ar is an (n, d) np.ndarray.

    Used in classic optimization and mcmc, for instance.
    """
    if not isinstance(arr, np.ndarray):
        raise ValueError("Please enter arr of shape (n, d)")
    elif arr.ndim != 2:
        raise ValueError("Please enter arr of shape (n, d)")



"""
Gaussian filtering and smoothing based on making intractable quantities
tractable through Taylor-method approximations, e.g. linearization.
"""
import numpy as np

from probnum.filtsmooth.gaussfiltsmooth.gaussfiltsmooth import *
from probnum.prob import RandomVariable
from probnum.prob.distributions import Normal
from probnum.filtsmooth.statespace import *


class ExtendedKalman(GaussFiltSmooth):
    """
    Factory method for Kalman filters.
    """

    def __new__(cls, dynamod, measmod, initrv, **kwargs):
        """ """
        if cls is ExtendedKalman:
            if _cont_disc(dynamod, measmod):
                return _ContDiscExtendedKalman(dynamod, measmod, initrv, **kwargs)
            if _disc_disc(dynamod, measmod):
                return _DiscDiscExtendedKalman(dynamod, measmod, initrv, **kwargs)
            else:
                errmsg = (
                    "Cannot instantiate Extended Kalman filter with "
                    "given dynamic model and measurement model."
                )
                raise ValueError(errmsg)
        else:
            return super().__new__(cls)


def _cont_disc(dynamod, measmod):
    """Check whether the state space model is continuous-discrete."""
    dyna_is_cont = issubclass(type(dynamod), ContinuousModel)
    meas_is_disc = issubclass(type(measmod), DiscreteModel)
    return dyna_is_cont and meas_is_disc


def _disc_disc(dynamod, measmod):
    """Check whether the state space model is discrete-discrete."""
    dyna_is_disc = issubclass(type(dynamod), DiscreteModel)
    meas_is_disc = issubclass(type(measmod), DiscreteModel)
    return dyna_is_disc and meas_is_disc


class _ContDiscExtendedKalman(ExtendedKalman):
    """
    Continuous-discrete extended Kalman filtering and smoothing.
    """

    def __init__(self, dynamod, measmod, initrv, **kwargs):
        if not issubclass(type(dynamod), LinearSDEModel):
            raise ValueError(
                "This implementation of ContDiscExtendedKalmanFilter "
                "requires a linear dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianModel):
            raise ValueError(
                "ContDiscExtendedKalmanFilter requires a Gaussian measurement model."
            )
        if "cke_nsteps" in kwargs.keys():
            self.cke_nsteps = kwargs["cke_nsteps"]
        else:
            self.cke_nsteps = 1
        super().__init__(dynamod, measmod, initrv)

    def predict(self, start, stop, randvar, **kwargs):
        """ """
        step = (stop - start) / self.cke_nsteps
        return self.dynamicmodel.chapmankolmogorov(start, stop, step, randvar, **kwargs)

    def update(self, time, randvar, data, **kwargs):
        """ """
        return _discrete_extkalman_update(
            time, randvar, data, self.measurementmodel, **kwargs
        )


class _DiscDiscExtendedKalman(ExtendedKalman):
    """
    """

    def __init__(self, dynamod, measmod, initrv, **kwargs):
        """
        Checks that dynamod and measmod are linear and moves on.
        """
        if not issubclass(type(dynamod), DiscreteGaussianModel):
            raise ValueError(
                "DiscDiscExtendedKalmanFilter requires " "a linear dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianModel):
            raise ValueError(
                "DiscDiscExtendedKalmanFilter requires " "a linear measurement model."
            )
        super().__init__(dynamod, measmod, initrv)

    def predict(self, start, stop, randvar, **kwargs):
        """ """
        mean, covar = randvar.mean(), randvar.cov()
        if np.isscalar(mean) and np.isscalar(covar):
            mean, covar = mean * np.ones(1), covar * np.eye(1)
        diffmat = self.dynamod.diffusionmatrix(start, **kwargs)
        jacob = self.dynamod.jacobian(start, mean, **kwargs)
        mpred = self.dynamod.dynamics(start, mean, **kwargs)
        crosscov = covar @ jacob.T
        cpred = jacob @ crosscov + diffmat
        return RandomVariable(distribution=Normal(mpred, cpred)), crosscov

    def update(self, time, randvar, data, **kwargs):
        """ """
        return _discrete_extkalman_update(
            time, randvar, data, self.measurementmodel, **kwargs
        )


def _discrete_extkalman_update(time, randvar, data, measmod, **kwargs):
    """
    """
    mpred, cpred = randvar.mean(), randvar.cov()
    if np.isscalar(mpred) and np.isscalar(cpred):
        mpred, cpred = mpred * np.ones(1), cpred * np.eye(1)
    jacob = measmod.jacobian(time, mpred, **kwargs)
    meascov = measmod.diffusionmatrix(time, **kwargs)
    meanest = measmod.dynamics(time, mpred, **kwargs)
    covest = jacob @ cpred @ jacob.T + meascov
    ccest = cpred @ jacob.T
    mean = mpred + ccest @ np.linalg.solve(covest, data - meanest)
    cov = cpred - ccest @ np.linalg.solve(covest.T, ccest.T)
    updated_rv = RandomVariable(distribution=Normal(mean, cov))
    return updated_rv, covest, ccest, meanest



"""
"""
import warnings
import numpy as np

from probnum.prob import RandomVariable
from probnum.prob.distributions import Normal
from probnum.diffeq import odesolver
from probnum.diffeq.odefiltsmooth.prior import ODEPrior
from probnum.filtsmooth import *


class GaussianIVPFilter(odesolver.ODESolver):
    """
    ODE solver that behaves like a Gaussian filter.


    This is based on continuous-discrete Gaussian filtering.

    Note: this is specific for IVPs and does not apply without
    further considerations to, e.g., BVPs.
    """

    def __init__(self, ivp, gaussfilt, steprl):
        """
        steprule : stepsize rule
        gaussfilt : gaussianfilter.GaussianFilter object,
            e.g. the return value of ivp_to_ukf(), ivp_to_ekf1().

        Notes
        -----
        * gaussfilt.dynamicmodel contains the prior,
        * gaussfilt.measurementmodel contains the information about the
        ODE right hand side function,
        * gaussfilt.initialdistribution contains the information about
        the initial values.
        """
        if not issubclass(type(gaussfilt.dynamicmodel), ODEPrior):
            raise ValueError("Please initialise a Gaussian filter with an ODEPrior")
        self.ivp = ivp
        self.gfilt = gaussfilt
        odesolver.ODESolver.__init__(self, steprl)

    def solve(self, firststep, nsteps=1, **kwargs):
        """
        Solves IVP and calibrates uncertainty according
        to Proposition 4 in Tronarp et al.

        Parameters
        ----------
        firststep : float
            First step for adaptive step size rule.
        nsteps : int, optional
            Number of inbetween steps for the filter. Default is 1.
        """

        ####### This function surely can use some code cleanup. #######

        current = self.gfilt.initialdistribution
        step = firststep
        ssqest, ct = 0.0, 0
        times, means, covars = [self.ivp.t0], [current.mean()], [current.cov()]
        while times[-1] < self.ivp.tmax:
            intermediate_step = float(step / nsteps)
            tm = times[-1]
            interms, intercs, interts = [], [], []
            for idx in range(nsteps):
                newtm = tm + intermediate_step
                current, __ = self.gfilt.predict(tm, newtm, current, **kwargs)
                interms.append(current.mean().copy())
                intercs.append(current.cov().copy())
                interts.append(newtm)
                tm = newtm
            predicted = current
            new_time = tm
            zero_data = 0.0
            current, covest, ccest, mnest = self.gfilt.update(
                new_time, predicted, zero_data, **kwargs
            )
            interms[-1] = current.mean().copy()
            intercs[-1] = current.cov().copy()
            errorest, ssq = self._estimate_error(current.mean(), ccest, covest, mnest)
            if self.steprule.is_accepted(step, errorest) is True:
                times.extend(interts)
                means.extend(interms)
                covars.extend(intercs)
                ct = ct + 1
                ssqest = ssqest + (ssq - ssqest) / ct
            else:
                current = RandomVariable(distribution=Normal(means[-1], covars[-1]))
            step = self._suggest_step(step, errorest)
        means, covars = self.undo_preconditioning(means, covars)
        return np.array(means), ssqest * np.array(covars), np.array(times)

    def odesmooth(self, means, covs, times, **kwargs):
        """
        Smoothes out the ODE-Filter output.

        Be careful about the preconditioning: the GaussFiltSmooth object
        only knows the state space with changed coordinates!

        Parameters
        ----------
        means
        covs

        Returns
        -------

        """
        means, covs = self.redo_preconditioning(means, covs)
        means, covs = self.gfilt.smooth(means, covs, times, **kwargs)
        return self.undo_preconditioning(means, covs)

    def undo_preconditioning(self, means, covs):
        """ """
        ipre = self.gfilt.dynamicmodel.invprecond
        newmeans = np.array([ipre @ mean for mean in means])
        newcovs = np.array([ipre @ cov @ ipre.T for cov in covs])
        return newmeans, newcovs

    def redo_preconditioning(self, means, covs):
        """ """
        pre = self.gfilt.dynamicmodel.precond
        newmeans = np.array([pre @ mean for mean in means])
        newcovs = np.array([pre @ cov @ pre.T for cov in covs])
        return newmeans, newcovs

    def _estimate_error(self, currmn, ccest, covest, mnest):
        """
        Error estimate.

        Estimates error as whitened residual using the
        residual as weight vector.

        THIS IS NOT A PERFECT ERROR ESTIMATE, but this is a question of
        research, not a question of implementation as of now.
        """
        std_like = np.linalg.cholesky(covest)
        whitened_res = np.linalg.solve(std_like, mnest)
        ssq = whitened_res @ whitened_res / mnest.size
        abserrors = np.abs(whitened_res)
        errorest = self._rel_and_abs_error(abserrors, currmn)
        return errorest, ssq

    def _rel_and_abs_error(self, abserrors, currmn):
        """
        Returns maximum of absolute and relative error.
        This way, both are guaranteed to be satisfied.
        """
        ordint, spatialdim = self.gfilt.dynamicmodel.ordint, self.ivp.ndim
        h0_1d = np.eye(ordint + 1)[:, 0].reshape((1, ordint + 1))
        projmat = np.kron(np.eye(spatialdim), h0_1d)
        weights = np.ones(len(abserrors))
        rel_error = (
            (abserrors / np.abs(projmat @ currmn)) @ weights / np.linalg.norm(weights)
        )
        abs_error = abserrors @ weights / np.linalg.norm(weights)
        return np.maximum(rel_error, abs_error)

    def _suggest_step(self, step, errorest):
        """
        Suggests step according to steprule and warns if
        step is extremely small.

        Raises
        ------
        RuntimeWarning
            If suggested step is smaller than :math:`10^{-15}`.
        """
        step = self.steprule.suggest(step, errorest)
        if step < 1e-15:
            warnmsg = "Stepsize is num. zero (%.1e)" % step
            warnings.warn(message=warnmsg, category=RuntimeWarning)
        return step

    @property
    def prior(self):
        """ """
        return self.gfilt.dynamicmodel



"""
Bayesian Quadrature.

This module provides routines to integrate functions through Bayesian quadrature, meaning a model over the integrand
is constructed in order to actively select evaluation points of the integrand to estimate the value of the integral.
Bayesian quadrature methods return a random variable with a distribution, specifying the belief about the true value of
the integral.
"""

import numpy as np
from probnum.quad.quadrature import Quadrature


def bayesquad(fun, fun0, bounds, nevals=None, type="vanilla", **kwargs):
    """
    One-dimensional Bayesian quadrature.

    Parameters
    ----------
    fun : function
        Function to be integrated.
    fun0 : RandomProcess
        Stochastic process modelling the function to be integrated.
    bounds : ndarray, shape=(2,)
        Lower and upper limit of integration.
    nevals : int
        Number of function evaluations.
    type : str
        Type of Bayesian quadrature to use. The available options are

        ====================  ===========
         vanilla              ``vanilla``
         WSABI                ``wsabi``
        ====================  ===========

    kwargs : optional
        Keyword arguments passed on.

    Returns
    -------
    F : RandomVariable
        The integral of ``func`` from ``a`` to ``b``.
    fun0 : RandomProcess
        Stochastic process modelling the function to be integrated after ``neval`` observations.
    info : dict
        Information on the performance of the method.

    References
    ----------

    See Also
    --------
    nbayesquad : N-dimensional Bayesian quadrature.
    """

    # Choose Method
    bqmethod = None
    if type == "vanilla":
        bqmethod = VanillaBayesianQuadrature()
    elif type == "wsabi":
        bqmethod = WSABIBayesianQuadrature()

    # Integrate
    F, fun0, info = bqmethod.integrate(
        fun=fun, fun0=fun0, nevals=nevals, domain=bounds, **kwargs
    )

    return F, fun0, info


def nbayesquad(fun, fun0, domain, nevals=None, type=None, **kwargs):
    """
    N-dimensional Bayesian quadrature.

    Parameters
    ----------
    fun : function
        Function to be integrated.
    fun0 : RandomProcess
        Stochastic process modelling the function to be integrated.
    domain : ndarray
        Domain of integration.
    nevals : int
        Number of function evaluations.
    type : str
        Type of Bayesian quadrature to use. The available options are

        ====================  ===========
         vanilla              ``vanilla``
         WASABI               ``wasabi``
        ====================  ===========

    kwargs : optional
        Keyword arguments passed on.

    Returns
    -------
    F : RandomVariable
        The integral of ``func`` on the domain.
    fun0 : RandomProcess
        Stochastic process modelling the function to be integrated after ``neval`` observations.
    info : dict
        Information on the performance of the method.

    References
    ----------

    See Also
    --------
    bayesquad : 1-dimensional Bayesian quadrature.
    """
    raise NotImplementedError


class BayesianQuadrature(Quadrature):
    """
    An abstract base class for Bayesian quadrature methods.

    This class is designed to be subclassed by implementations of Bayesian quadrature with an :meth:`integrate` method.
    """

    def __init__(self):
        super().__init__()

    def integrate(self, fun, fun0, domain, nevals, **kwargs):
        """
        Integrate the function ``fun``.

        Parameters
        ----------
        fun : function
            Function to be integrated.
        fun0 : RandomProcess
            Stochastic process modelling function to be integrated.
        domain : ndarray, shape=()
            Domain to integrate over.
        nevals : int
            Number of function evaluations.
        kwargs

        Returns
        -------

        """
        raise NotImplementedError


class VanillaBayesianQuadrature(BayesianQuadrature):
    """
    Vanilla Bayesian quadrature in 1D.
    """

    def __init__(self):
        super().__init__()

    def integrate(self, fun, fun0, domain, nevals, **kwargs):
        """
        Integrate the function ``fun``.

        Parameters
        ----------
        fun : function
            Function to be integrated.
        fun0 : RandomProcess
            Stochastic process modelling function to be integrated.
        domain : ndarray, shape=()
            Domain to integrate over.
        nevals : int
            Number of function evaluations.

        Returns
        -------

        """

        # Initialization
        F = None

        # Iteration
        for i in range(nevals):
            # Predictive Distribution
            fun_pred = None

            # Observation

            # Update Distribution
            self.fun0 = None

        # Information on result
        info = {"model_fit_diagnostic": None}

        return F, self.fun0, info


class WSABIBayesianQuadrature(BayesianQuadrature):
    """
    Warped Sequential Active Bayesian Integration (WSABI).
    """

    def __init__(self):
        super().__init__()

    def integrate(self, fun, fun0, domain, nevals, **kwargs):
        """
        Integrate the function ``fun``.

        Parameters
        ----------
        fun : function
            Function to be integrated.
        fun0 : RandomProcess
            Stochastic process modelling function to be integrated.
        domain : ndarray, shape=()
            Domain to integrate over.
        nevals : int
            Number of function evaluations.

        Returns
        -------

        """

        raise NotImplementedError



"""
Convenience functions for Gaussian filtering and smoothing.

We support the following methods:
    - ekf0: Extended Kalman filtering based on a zero-th order Taylor
        approximation [1]_, [2]_, [3]_. Also known as "PFOS".
    - ekf1: Extended Kalman filtering [3]_.
    - ukf: Unscented Kalman filtering [3]_.
    - eks0: Extended Kalman smoothing based on a zero-th order Taylor
        approximation [4]_.
    - eks1: Extended Kalman smoothing [4]_.
    - uks: Unscented Kalman smoothing.

References
----------
.. [1] https://arxiv.org/pdf/1610.05261.pdf
.. [2] https://arxiv.org/abs/1807.09737
.. [3] https://arxiv.org/abs/1810.03440
.. [4] https://arxiv.org/pdf/2004.00623.pdf
"""

from probnum.diffeq import steprule
from probnum.diffeq.odefiltsmooth import prior, ivp2filter
from probnum.diffeq.odefiltsmooth import GaussianIVPFilter


def probsolve_ivp(
    ivp,
    method="ekf0",
    which_prior="ibm1",
    tol=None,
    step=None,
    firststep=None,
    precond_step=1.0,
    nsteps=1,
    **kwargs
):
    """
    Solve initial value problem with Gaussian filtering and smoothing.

    Numerically computes a Gauss-Markov process which solves numerically
    the initial value problem (IVP) based on a system of first order
    ordinary differential equations (ODEs)

    .. math:: \\dot x(t) = f(t, x(t)), \\quad x(t_0) = x_0,
        \\quad t \\in [t_0, T]

    by regarding it as a (nonlinear) Gaussian filtering (and smoothing)
    problem [3]_. For some configurations it recovers certain multistep
    methods [1]_.
    Convergence rates of filtering [2]_ and smoothing [4]_ are
    comparable to those of methods of Runge-Kutta type.


    This function turns a prior-string into an :class:`ODEPrior`, a
    method-string into a :class:`GaussianSmoother`, creates a
    :class:`GaussianIVPSmoother` object and calls the :meth:`solve()` method. For
    advanced usage we recommend to do this process manually which
    enables advanced methods of tuning the algorithm.

    This function supports the methods:
    extended Kalman filtering based on a zero-th order Taylor
    approximation (EKF0),
    extended Kalman filtering (EKF1),
    unscented Kalman filtering (UKF),
    extended Kalman smoothing based on a zero-th order Taylor
    approximation (EKS0),
    extended Kalman smoothing (EKS1), and
    unscented Kalman smoothing (UKS).

    Arguments
    ---------
    ivp : IVP
        Initial value problem to be solved.
    step : float
        Step size :math:`h` of the solver. This defines the
        discretisation mesh as each proposed step is equal to :math:`h`
        and all proposed steps are accepted.
        Only one of out of ``step`` and ``tol`` is set.
    tol : float
        Tolerance :math:`\\varepsilon` of the adaptive step scheme.
        We implement the scheme proposed by Schober et al., accepting a
        step if the absolute as well as the relative error estimate are
        smaller than the tolerance,
        :math:`\\max\\{e, e / |y|\\} \\leq \\varepsilon`.
        Only one of out of ``step`` and ``tol`` is set.
    which_prior : str, optional
        Which prior is to be used. Default is an IBM(1), further support
        for IBM(:math:`q`), IOUP(:math:`q`), Matern(:math:`q+1/2`),
        :math:`q\\in\\{1, 2, 3, 4\\}` is provided. The available
        options are

        ======================  ========================================
         IBM(:math:`q`)         ``'ibm1'``, ``'ibm2'``, ``'ibm3'``,
                                ``'ibm4'``
         IOUP(:math:`q`)        ``'ioup1'``, ``'ioup2'``, ``'ioup3'``,
                                ``'ioup4'``
         Matern(:math:`q+0.5`)  ``'matern32'``, ``'matern52'``,
                                ``'matern72'``, ``'matern92'``
        ======================  ========================================

        The type of prior relates to prior assumptions about the
        derivative of the solution. The IBM(:math:`q`) prior leads to a
        :math:`q`-th order method that is recommended if little to no
        prior information about the solution is available. On the other
        hand, if the :math:`q`-th derivative is expected to regress to
        zero, an IOUP(:math:`q`) prior might be suitable.
    method : str, optional
        Which method is to be used. Default is ``ekf0`` which is the
        method proposed by Schober et al.. The available
        options are

        ================================================  ==============
         Extended Kalman filtering/smoothing (0th order)  ``'ekf0'``,
                                                          ``'eks0'``
         Extended Kalman filtering/smoothing (1st order)  ``'ekf1'``,
                                                          ``'eks1'``
         Unscented Kalman filtering/smoothing             ``'ukf'``,
                                                          ``'uks'``
        ================================================  ==============

        First order extended Kalman filtering and smoothing methods
        require Jacobians of the RHS-vector field of the IVP. The
        uncertainty estimates as returned by EKF1/S1 and UKF/S appear to
        be more reliable than those of EKF0/S0. The latter is more
        stable when it comes to very small steps.

    firststep : float, optional
        First suggested step :math:`h_0` for adaptive step size scheme.
        Default is None which lets the solver start with the suggestion
        :math:`h_0 = T - t_0`. For low accuracy it might be more
        efficient to start out with smaller :math:`h_0` so that the
        first acceptance occurs earlier.

    precond_step : float, optional
        Expected average step size, used for preconditioning.
        See :class:`ODEPrior` for details.
        Default is ``precond_step=1.0`` which amounts to no
        preconditioning. If constant step size, precond_step is
        overwritten with the actual step size to provide optimal
        preconditioning.

    nsteps : int, optional
        Number of intermediate steps in between ODE right-hand side
        evaluation. Default is ``nsteps=1``. Choosing ``nsteps > 1``
        will result in solution estimates in between points which
        will enable for instance uncertainty estimates in these regions.

    Returns
    -------
    means : np.ndarray, shape=(N, d(q+1))
        Mean vector of the solution at times :math:`t_1, ..., t_N`.
        The elements are ordered as
        ``(m_1, m_1\', m_1\'\', ..., m_2, m_2\', ...)``
        where ``m_1`` is the estimate of the first coordinate of the
        solution, ``m_1\'`` is the estimate of the derivative of the
        first coordinate, and so on.
    covs : np.ndarray, shape=(N, d(q+1), d(q+1))
        Covariance matrices of the solution at times
        :math:`t_1, ..., t_N`. The ordering reflects the ordering of
        the mean vector.
    times : np.ndarray, shape=(N,)
        Mesh used by the solver to compute the solution.
        It includes the initial time :math:`t_0` but not necessarily the
        final time :math:`T`.

    See Also
    --------
    GaussianIVPFilter : Solve IVPs with Gaussian filtering.
    GaussianIVPSmoother : Solve IVPs with Gaussian smoothing.

    References
    ----------
    .. [1] Schober, M., Särkkä, S. and Hennig, P..
        A probabilistic model for the numerical solution of initial
        value problems.
        Statistics and Computing, 2019.
    .. [2] Kersting, H., Sullivan, T.J., and Hennig, P..
        Convergence rates of Gaussian ODE filters.
        2019.
    .. [3] Tronarp, F., Kersting, H., Särkkä, S., and Hennig, P..
        Probabilistic solutions to ordinary differential equations as
        non-linear Bayesian filtering: a new perspective.
        Statistics and Computing, 2019.
    .. [4] Tronarp, F., Särkkä, S., and Hennig, P..
        Bayesian ODE solvers: the maximum a posteriori estimate.
        2019.


    Examples
    --------
    >>> from probnum.diffeq import logistic, probsolve_ivp
    >>> from probnum.prob import RandomVariable, Dirac, Normal
    >>> initrv = RandomVariable(distribution=Dirac(0.15))
    >>> ivp = logistic(timespan=[0., 1.5], initrv=initrv, params=(4, 1))
    >>> means, covs, times = probsolve_ivp(ivp, method="ekf0", step=0.1)
    >>> print(means)
    [[0.15       0.51      ]
     [0.2076198  0.642396  ]
     [0.27932997 0.79180747]
     [0.3649165  0.91992313]
     [0.46054129 0.9925726 ]
     [0.55945475 0.98569653]
     [0.65374523 0.90011316]
     [0.73686744 0.76233098]
     [0.8053776  0.60787222]
     [0.85895587 0.4636933 ]
     [0.89928283 0.34284592]
     [0.92882899 0.24807715]
     [0.95007559 0.17685497]
     [0.96515825 0.12479825]
     [0.97577054 0.08744746]
     [0.9831919  0.06097975]]

    >>> initrv = RandomVariable(distribution=Dirac(0.15))
    >>> ivp = logistic(timespan=[0., 1.5], initrv=initrv, params=(4, 1))
    >>> means, covs, times = probsolve_ivp(ivp, method="eks1", which_prior="ioup3", step=0.1)
    >>> print(means)
    [[  0.15         0.51         1.428        1.9176    ]
     [  0.20795795   0.65884674   2.00211064  -6.59817856]
     [  0.28228416   0.81039925   1.10201443   1.99947952]
     [  0.36926      0.93163093   1.15809878 -10.29411145]
     [  0.46646494   0.99550339   0.17201064  -6.25619529]
     [  0.5658788    0.98264107  -0.45761145  -8.49165628]
     [  0.66048505   0.89697823  -1.1844693   -3.69769413]
     [  0.74369892   0.76244437  -1.47053431  -1.49540782]
     [  0.81237394   0.60969196  -1.53808527   1.57069856]
     [  0.86592791   0.46438819  -1.35517427   2.37556924]
     [  0.90598293   0.34071182  -1.11171014   2.83875749]
     [  0.9349573    0.24324862  -0.84433565   2.40611949]
     [  0.95544749   0.17027033  -0.62177491   2.00601395]
     [  0.96968754   0.11757447  -0.44061846   1.46774923]
     [  0.97947631   0.08040988  -0.30900324   1.07108871]
     [  0.98614541   0.05465059  -0.20824105   0.94815346]]
    """
    solver, firststep = _create_solver_object(
        ivp, method, which_prior, tol, step, firststep, precond_step, nsteps, **kwargs
    )
    means, covs, times = solver.solve(firststep=firststep, nsteps=nsteps, **kwargs)
    if method in ["eks0", "eks1", "uks"]:
        means, covs = solver.odesmooth(means, covs, times, **kwargs)
    return means, covs, times


def _create_solver_object(
    ivp, method, which_prior, tol, step, firststep, precond_step, nsteps, **kwargs
):
    """Create the solver object that is used."""
    _check_step_tol(step, tol)
    _check_method(method)
    if step is not None:
        precond_step = step
    precond_step = precond_step / float(nsteps)
    _prior = _string2prior(ivp, which_prior, precond_step, **kwargs)
    if tol is not None:
        stprl = _step2steprule_adap(tol, _prior)
        if firststep is None:
            firststep = ivp.tmax - ivp.t0
    else:
        stprl = _step2steprule_const(step)
        firststep = step
    gfilt = _string2filter(ivp, _prior, method, **kwargs)
    return GaussianIVPFilter(ivp, gfilt, stprl), firststep


def _check_step_tol(step, tol):
    """ """
    both_none = tol is None and step is None
    both_not_none = tol is not None and step is not None
    if both_none or both_not_none:
        errormsg = "Please specify either a tolerance or a step size."
        raise ValueError(errormsg)


def _check_method(method):
    """ """
    if method not in ["ekf0", "ekf1", "ukf", "eks0", "eks1", "uks"]:
        raise ValueError("Method not supported.")


def _string2prior(ivp, which_prior, precond_step, **kwargs):
    """
    """
    ibm_family = ["ibm1", "ibm2", "ibm3", "ibm4"]
    ioup_family = ["ioup1", "ioup2", "ioup3", "ioup4"]
    matern_family = ["matern32", "matern52", "matern72", "matern92"]
    if which_prior in ibm_family:
        return _string2ibm(ivp, which_prior, precond_step, **kwargs)
    elif which_prior in ioup_family:
        return _string2ioup(ivp, which_prior, precond_step, **kwargs)
    elif which_prior in matern_family:
        return _string2matern(ivp, which_prior, precond_step, **kwargs)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2ibm(ivp, which_prior, precond_step, **kwargs):
    """
    """
    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if which_prior == "ibm1":
        return prior.IBM(1, ivp.ndim, diffconst, precond_step)
    elif which_prior == "ibm2":
        return prior.IBM(2, ivp.ndim, diffconst, precond_step)
    elif which_prior == "ibm3":
        return prior.IBM(3, ivp.ndim, diffconst, precond_step)
    elif which_prior == "ibm4":
        return prior.IBM(4, ivp.ndim, diffconst, precond_step)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2ioup(_ivp, _which_prior, precond_step, **kwargs):
    """
    """
    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if "driftspeed" in kwargs.keys():
        driftspeed = kwargs["driftspeed"]
    else:
        driftspeed = 1.0
    if _which_prior == "ioup1":
        return prior.IOUP(1, _ivp.ndim, driftspeed, diffconst, precond_step)
    elif _which_prior == "ioup2":
        return prior.IOUP(2, _ivp.ndim, driftspeed, diffconst, precond_step)
    elif _which_prior == "ioup3":
        return prior.IOUP(3, _ivp.ndim, driftspeed, diffconst, precond_step)
    elif _which_prior == "ioup4":
        return prior.IOUP(4, _ivp.ndim, driftspeed, diffconst, precond_step)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2matern(ivp, which_prior, precond_step, **kwargs):
    """
    """
    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if "lengthscale" in kwargs.keys():
        lengthscale = kwargs["lengthscale"]
    else:
        lengthscale = 1.0
    if which_prior == "matern32":
        return prior.Matern(1, ivp.ndim, lengthscale, diffconst, precond_step)
    elif which_prior == "matern52":
        return prior.Matern(2, ivp.ndim, lengthscale, diffconst, precond_step)
    elif which_prior == "matern72":
        return prior.Matern(3, ivp.ndim, lengthscale, diffconst, precond_step)
    elif which_prior == "matern92":
        return prior.Matern(4, ivp.ndim, lengthscale, diffconst, precond_step)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2filter(_ivp, _prior, _method, **kwargs):
    """
    """
    if "evlvar" in kwargs.keys():
        evlvar = kwargs["evlvar"]
    else:
        evlvar = 0.0
    if _method == "ekf0" or _method == "eks0":
        return ivp2filter.ivp2ekf0(_ivp, _prior, evlvar)
    elif _method == "ekf1" or _method == "eks1":
        return ivp2filter.ivp2ekf1(_ivp, _prior, evlvar)
    elif _method == "ukf" or _method == "uks":
        return ivp2filter.ivp2ukf(_ivp, _prior, evlvar)
    else:
        raise ValueError("Type of filter not supported.")


def _step2steprule_const(stp):
    """
    """
    return steprule.ConstantSteps(stp)


def _step2steprule_adap(_tol, _prior, **kwargs):
    """
    """
    convrate = _prior.ordint + 1
    return steprule.AdaptiveSteps(_tol, convrate, **kwargs)



"""
Normal distribution.

This module implements the Gaussian distribution in its univariate,
multi-variate, matrix-variate and operator-variate forms.
"""
import operator
import warnings

import numpy as np
import scipy.stats
import scipy.sparse
import scipy._lib._util

from probnum.linalg import linops
from probnum.prob.distributions.distribution import Distribution
from probnum.prob.distributions.dirac import Dirac


class Normal(Distribution):
    """
    The normal distribution.

    The Gaussian distribution is ubiquitous in probability theory, since
    it is the final and stable or equilibrium distribution to which
    other distributions gravitate under a wide variety of smooth
    operations, e.g., convolutions and stochastic transformations.
    One example of this is the central limit theorem. The Gaussian
    distribution is also attractive from a numerical point of view as it
    is maintained through many transformations (e.g. it is stable).

    Parameters
    ----------
    mean : float or array-like or LinearOperator
        Mean of the normal distribution.

    cov : float or array-like or LinearOperator
        (Co-)variance of the normal distribution.

    random_state : None or int or :class:`~numpy.random.RandomState` instance, optional
        This parameter defines the RandomState object to
        use for drawing realizations from this
        distribution. Think of it like a random seed.
        If None (or np.random), the global
        np.random state is used. If integer, it is used to
        seed the local
        :class:`~numpy.random.RandomState` instance.
        Default is None.

    See Also
    --------
    Distribution : Class representing general probability distributions.

    Examples
    --------
    >>> from probnum.prob import Normal
    >>> N = Normal(mean=0.5, cov=1.0)
    >>> N.parameters
    {'mean': 0.5, 'cov': 1.0}
    """

    def __new__(cls, mean=0.0, cov=1.0, random_state=None):
        """
        Factory method for normal subclasses.

        Checks shape/type of mean and cov and returns the corresponding
        type of Normal distribution:
            * _UnivariateNormal
            * _MultivariateNormal
            * _MatrixvariateNormal
            * _SymmetricKroneckerIdenticalFactorsNormal
            * _OperatorvariateNormal
        If neither applies, a ValueError is raised.
        """
        if cls is Normal:
            if _both_are_univariate(mean, cov):
                return _UnivariateNormal(mean, cov, random_state)
            elif _both_are_multivariate(mean, cov):
                return _MultivariateNormal(mean, cov, random_state)
            elif _both_are_matrixvariate(mean, cov):
                return _MatrixvariateNormal(mean, cov, random_state)
            elif _both_are_symmkronidfactors(mean, cov):
                return _SymmetricKroneckerIdenticalFactorsNormal(
                    mean, cov, random_state
                )
            elif _both_are_operatorvariate(mean, cov):
                return _OperatorvariateNormal(mean, cov, random_state)
            else:
                errmsg = (
                    "Cannot instantiate normal distribution with mean of "
                    + "type {} and ".format(mean.__class__.__name__)
                    + "kernels of "
                    + "type {}.".format(cov.__class__.__name__)
                )
                raise ValueError(errmsg)
        else:
            return super().__new__(cls)

    def __init__(self, mean=0.0, cov=1.0, random_state=None):
        super().__init__(
            parameters={"mean": mean, "cov": cov},
            dtype=float,
            random_state=random_state,
        )

    def mean(self):
        return self.parameters["mean"]

    def cov(self):
        return self.parameters["cov"]

    def var(self):
        raise NotImplementedError

    # Binary arithmetic

    def __add__(self, other):
        """
        Addition of Gaussian random variables.
        """
        if isinstance(other, Dirac):
            return Normal(
                mean=self.mean() + other.mean(),
                cov=self.cov(),
                random_state=self.random_state,
            )
        elif isinstance(other, type(self)):
            if self.random_state is not None and other.random_state is not None:
                warnings.warn(
                    "When adding random variables with set random states only the first is preserved."
                )
            try:
                return Normal(
                    mean=self.mean() + other.mean(),
                    cov=self.cov() + other.cov(),
                    random_state=self.random_state,
                )
            except ValueError:
                return NotImplemented
        else:
            return NotImplemented

    def __sub__(self, other):
        """
        Subtraction of Gaussian random variables.
        """
        if isinstance(other, Dirac):
            return self + (-other)
        elif isinstance(other, type(self)):
            if self.random_state is not None and other.random_state is not None:
                warnings.warn(
                    "When adding random variables with set random states only the first is preserved."
                )
            try:
                return Normal(
                    mean=self.mean() - other.mean(),
                    cov=self.cov() + other.cov(),
                    random_state=self.random_state,
                )
            except ValueError:
                return NotImplemented
        else:
            return NotImplemented

    def __mul__(self, other):
        """
        Only works if the other dist. is a Dirac.
        """
        if isinstance(other, Dirac):
            delta = other.mean()
            if delta == 0:
                return Dirac(support=0 * self.mean(), random_state=self.random_state)
            else:
                return Normal(
                    mean=self.mean() * delta,
                    cov=self.cov() * delta ** 2,
                    random_state=self.random_state,
                )
        else:
            return NotImplemented

    def __truediv__(self, other):
        """
        Only works if the other dist. is a Dirac.
        """
        if other == 0:
            raise ZeroDivisionError("Division by zero not supported.")
        else:
            if isinstance(other, Dirac):
                return self * operator.inv(other)
            else:
                return NotImplemented

    def __pow__(self, power, modulo=None):
        return NotImplemented

    # Binary arithmetic with reflected operands

    def __radd__(self, other):
        """
        Only works if the other dist. is a Dirac.
        """
        if isinstance(other, Dirac):
            return self + other
        else:
            return NotImplemented

    def __rsub__(self, other):
        """
        Only works if the other dist. is a Dirac.
        """
        if isinstance(other, Dirac):
            return operator.neg(self) + other
        else:
            return NotImplemented

    def __rmul__(self, other):
        """
        Only works if the other dist. is a Dirac.
        """
        if isinstance(other, Dirac):
            return self * other
        else:
            return NotImplemented

    def __rmatmul__(self, other):
        """
        Only works if the other dist. is a Dirac.
        """
        if isinstance(other, Dirac):
            delta = other.mean()
            newmean = delta @ self.mean()
            newcov = delta @ (self.cov() @ delta.transpose())
            return Normal(mean=newmean, cov=newcov, random_state=self.random_state)
        return NotImplemented

    def __rtruediv__(self, other):
        """
        Only works if the other dist. is a Dirac.
        """
        if isinstance(other, Dirac):
            return operator.inv(self) * other
        else:
            return NotImplemented

    def __rpow__(self, power, modulo=None):
        return NotImplemented

    # Unary arithmetic

    def __neg__(self):
        """
        Negation of r.v.
        """
        try:
            return Normal(
                mean=-self.mean(), cov=self.cov(), random_state=self.random_state
            )
        except Exception:
            return NotImplemented

    def __pos__(self):
        try:
            return Normal(
                mean=operator.pos(self.mean()),
                cov=self.cov(),
                random_state=self.random_state,
            )
        except Exception:
            return NotImplemented

    def __abs__(self):
        try:
            # todo: add absolute moments of normal (see: https://arxiv.org/pdf/1209.4340.pdf)
            return Distribution(
                parameters={},
                sample=lambda size: operator.abs(self.sample(size=size)),
                random_state=self.random_state,
            )
        except Exception:
            return NotImplemented

    def __invert__(self):
        try:
            return Distribution(
                parameters={},
                sample=lambda size: operator.abs(self.sample(size=size)),
                random_state=self.random_state,
            )
        except Exception:
            return NotImplemented


def _both_are_univariate(mean, cov):
    """
    Checks whether mean and kernels correspond to the
    UNIVARIATE normal distribution.
    """
    both_are_scalars = np.isscalar(mean) and np.isscalar(cov)
    mean_shape_dim1 = np.shape(mean) in [(1, 1), (1,), ()]
    cov_shape_dim1 = np.shape(cov) in [(1, 1), (1,), ()]
    both_in_dim1shapes = mean_shape_dim1 and cov_shape_dim1
    return both_are_scalars or both_in_dim1shapes


def _both_are_multivariate(mean, cov):
    """
    Checks whether mean and kernels correspond to the
    MULTI- or MATRIXVARIATE normal distribution.
    """
    mean_is_multivar = isinstance(mean, (np.ndarray, scipy.sparse.spmatrix,))
    cov_is_multivar = isinstance(cov, (np.ndarray, scipy.sparse.spmatrix,))
    return mean_is_multivar and cov_is_multivar and len(mean.shape) == 1


def _both_are_matrixvariate(mean, cov):
    """
    Checks whether mean and kernels correspond to the
    MULTI- or MATRIXVARIATE normal distribution.
    """
    mean_is_multivar = isinstance(mean, (np.ndarray, scipy.sparse.spmatrix,))
    cov_is_multivar = isinstance(cov, (np.ndarray, scipy.sparse.spmatrix,))
    return mean_is_multivar and cov_is_multivar and len(mean.shape) > 1


def _both_are_symmkronidfactors(mean, cov):
    """
    Checks whether mean OR (!) kernels correspond to the
    OPERATORVARIATE normal distribution.
    """
    mean_is_opvariate = isinstance(mean, scipy.sparse.linalg.LinearOperator)
    cov_is_opvariate = isinstance(cov, scipy.sparse.linalg.LinearOperator)
    if mean_is_opvariate or cov_is_opvariate:
        return isinstance(cov, linops.SymmetricKronecker) and cov._ABequal
    else:
        return False


def _both_are_operatorvariate(mean, cov):
    """
    Checks whether mean OR (!) kernels correspond to the
    OPERATORVARIATE normal distribution.
    """
    mean_is_opvariate = isinstance(mean, scipy.sparse.linalg.LinearOperator)
    cov_is_opvariate = isinstance(cov, scipy.sparse.linalg.LinearOperator)
    return mean_is_opvariate or cov_is_opvariate


class _UnivariateNormal(Normal):
    """
    The univariate normal distribution.
    """

    def __init__(self, mean=0.0, cov=1.0, random_state=None):
        super().__init__(mean=float(mean), cov=float(cov), random_state=random_state)

    def var(self):
        return self.cov()

    def pdf(self, x):
        return scipy.stats.norm.pdf(x, loc=self.mean(), scale=self.std())

    def logpdf(self, x):
        return scipy.stats.norm.logpdf(x, loc=self.mean(), scale=self.std())

    def cdf(self, x):
        return scipy.stats.norm.cdf(x, loc=self.mean(), scale=self.std())

    def logcdf(self, x):
        return scipy.stats.norm.logcdf(x, loc=self.mean(), scale=self.std())

    def sample(self, size=()):
        return scipy.stats.norm.rvs(
            loc=self.mean(), scale=self.std(), size=size, random_state=self.random_state
        )

    def reshape(self, newshape):
        if np.prod(newshape) != 1:
            raise ValueError(
                f"Cannot reshape distribution with shape {self.shape} into shape {newshape}."
            )
        self.parameters["mean"] = np.reshape(self.parameters["mean"], newshape=newshape)
        self.parameters["cov"] = np.reshape(self.parameters["cov"], newshape=newshape)
        self._shape = newshape


class _MultivariateNormal(Normal):
    """
    The multivariate normal distribution.
    """

    def __init__(self, mean, cov, random_state=None):
        """
        Checks if mean and kernels have matching shapes before
        initialising.
        """
        meandim = np.prod(mean.shape)
        if len(cov.shape) != 2:
            raise ValueError("Covariance must be a 2D matrix " "or linear operator.")
        if meandim != cov.shape[0] or meandim != cov.shape[1]:
            raise ValueError(
                "Shape mismatch of mean and kernels. Total "
                "number of elements of the mean must match the "
                "first and second dimension of the kernels."
            )
        super().__init__(mean=mean, cov=cov, random_state=random_state)

    def var(self):
        return np.diag(self.cov())

    def pdf(self, x):
        return scipy.stats.multivariate_normal.pdf(x, mean=self.mean(), cov=self.cov())

    def logpdf(self, x):
        return scipy.stats.multivariate_normal.logpdf(
            x, mean=self.mean(), cov=self.cov()
        )

    def cdf(self, x):
        return scipy.stats.multivariate_normal.cdf(x, mean=self.mean(), cov=self.cov())

    def logcdf(self, x):
        return scipy.stats.multivariate_normal.logcdf(
            x, mean=self.mean(), cov=self.cov()
        )

    def sample(self, size=()):
        return scipy.stats.multivariate_normal.rvs(
            mean=self.mean(), cov=self.cov(), size=size, random_state=self.random_state
        )

    def reshape(self, newshape):
        raise NotImplementedError

    # Arithmetic Operations

    def __matmul__(self, other):
        """
        Only works if other is a Dirac, which implies
        matrix-multiplication with some constant.
        """
        if isinstance(other, Dirac):
            delta = other.mean()
            newmean = self.mean() @ delta
            newcov = delta.T @ (self.cov() @ delta)
            if np.isscalar(newmean) and np.isscalar(newcov):
                return _UnivariateNormal(
                    mean=newmean, cov=newcov, random_state=self.random_state
                )
            else:
                return _MultivariateNormal(
                    mean=newmean, cov=newcov, random_state=self.random_state
                )
        else:
            return NotImplemented

    def __rmatmul__(self, other):
        """
        Only works if other is a Dirac, which implies
        matrix-multiplication with some constant.
        """
        if isinstance(other, Dirac):
            delta = other.mean()
            newmean = delta @ self.mean()
            newcov = delta @ (self.cov() @ delta.T)
            if np.isscalar(newmean) and np.isscalar(newcov):
                return _UnivariateNormal(
                    mean=newmean, cov=newcov, random_state=self.random_state
                )
            else:
                return _MultivariateNormal(
                    mean=newmean, cov=newcov, random_state=self.random_state
                )
        else:
            return NotImplemented


class _MatrixvariateNormal(Normal):
    """
    The matrix-variate normal distribution.
    """

    def __init__(self, mean, cov, random_state=None):
        """
        Checks if mean and kernels have matching shapes before
        initialising.
        """
        _mean_dim = np.prod(mean.shape)
        if len(cov.shape) != 2:
            raise ValueError("Covariance must be a 2D matrix.")
        if _mean_dim != cov.shape[0] or _mean_dim != cov.shape[1]:
            raise ValueError(
                "Shape mismatch of mean and kernels. Total "
                "number of elements of the mean must match the "
                "first and second dimension of the kernels."
            )
        super().__init__(mean=mean, cov=cov, random_state=random_state)

    def var(self):
        return np.diag(self.cov())

    def pdf(self, x):
        pdf_ravelled = scipy.stats.multivariate_normal.pdf(
            x.ravel(), mean=self.mean().ravel(), cov=self.cov()
        )
        return pdf_ravelled.reshape(newshape=self.shape)

    def logpdf(self, x):
        raise NotImplementedError

    def cdf(self, x):
        raise NotImplementedError

    def logcdf(self, x):
        raise NotImplementedError

    def sample(self, size=()):
        ravelled = scipy.stats.multivariate_normal.rvs(
            mean=self.mean().ravel(),
            cov=self.cov(),
            size=size,
            random_state=self.random_state,
        )
        return ravelled.reshape(self.shape)

    def reshape(self, newshape):
        if np.prod(newshape) != np.prod(self.shape):
            raise ValueError(
                f"Cannot reshape distribution with shape {self.shape} into shape {newshape}."
            )
        self.parameters["mean"] = np.reshape(self.parameters["mean"], newshape=newshape)
        self.parameters["cov"] = np.reshape(self.parameters["cov"], newshape=newshape)
        self._shape = newshape

    # Arithmetic Operations
    # TODO: implement special rules for matrix-variate RVs and Kronecker
    #  structured covariances (see e.g. p.64 Thm. 2.3.10 of Gupta:
    #  Matrix-variate Distributions)

    def __matmul__(self, other):
        if isinstance(other, Dirac):
            delta = other.mean()
            raise NotImplementedError
        return NotImplemented


class _OperatorvariateNormal(Normal):
    """
    A normal distribution over finite dimensional linear operators.
    """

    def __init__(self, mean, cov, random_state=None):
        """
        Checks shapes of mean and cov depending on the type
        of operator that cov is before initialising.
        """
        self._mean_dim = np.prod(mean.shape)
        if isinstance(cov, linops.Kronecker):
            _check_shapes_if_kronecker(mean, cov)
        elif isinstance(cov, linops.SymmetricKronecker):
            _check_shapes_if_symmetric_kronecker(mean, cov)
        elif self._mean_dim != cov.shape[0] or self._mean_dim != cov.shape[1]:
            raise ValueError("Shape mismatch of mean and kernels.")
        super().__init__(mean=mean, cov=cov, random_state=random_state)

    def var(self):
        return linops.Diagonal(Op=self.cov())

    def _params_todense(self):
        """Returns the mean and kernels of a distribution as dense matrices."""
        if isinstance(self.mean(), linops.LinearOperator):
            mean = self.mean().todense()
        else:
            mean = self.mean()
        if isinstance(self.cov(), linops.LinearOperator):
            cov = self.cov().todense()
        else:
            cov = self.cov()
        return mean, cov

    def pdf(self, x):
        raise NotImplementedError

    def logpdf(self, x):
        raise NotImplementedError

    def cdf(self, x):
        raise NotImplementedError

    def logcdf(self, x):
        raise NotImplementedError

    def sample(self, size=()):
        mean, cov = self._params_todense()
        samples_ravelled = scipy.stats.multivariate_normal.rvs(
            mean=mean.ravel(), cov=cov, size=size, random_state=self.random_state
        )
        return samples_ravelled.reshape(samples_ravelled.shape[:-1] + self.mean().shape)

    def reshape(self, newshape):
        raise NotImplementedError

    # Arithmetic Operations

    # TODO: implement special rules for matrix-variate RVs and Kronecker structured covariances
    #  (see e.g. p.64 Thm. 2.3.10 of Gupta: Matrix-variate Distributions)
    def __matmul__(self, other):
        if isinstance(other, Dirac):
            othermean = other.mean()
            delta = linops.Kronecker(linops.Identity(othermean.shape[0]), othermean)
            return Normal(
                mean=self.mean() @ othermean,
                cov=delta.T @ (self.cov() @ delta),
                random_state=self.random_state,
            )
        return NotImplemented


def _check_shapes_if_kronecker(mean, cov):
    """
    If mean has dimension (m x n) then kernels factors must be
    (m x m) and (n x n)
    """
    m, n = mean.shape
    if (
        m != cov.A.shape[0]
        or m != cov.A.shape[1]
        or n != cov.B.shape[0]
        or n != cov.B.shape[1]
    ):
        raise ValueError(
            "Kronecker structured kernels must have "
            "factors with the same shape as the mean."
        )


class _SymmetricKroneckerIdenticalFactorsNormal(_OperatorvariateNormal):
    """
    Normal distribution with symmetric Kronecker structured
    kernels with identical factors V (x)_s V.
    """

    def __init__(self, mean, cov, random_state=None):
        _check_shapes_if_symmetric_kronecker(mean, cov)
        self._n = mean.shape[1]
        super().__init__(mean=mean, cov=cov, random_state=random_state)

    def sample(self, size=()):

        # Draw standard normal samples
        if np.isscalar(size):
            size = [size]
        size_sample = [self._n * self._n] + list(size)
        stdnormal_samples = scipy.stats.norm.rvs(
            size=size_sample, random_state=self.random_state
        )

        # Cholesky decomposition
        eps = 10 ** -12  # damping needed to avoid negative definite covariances
        cholA = scipy.linalg.cholesky(
            self.cov().A.todense() + eps * np.eye(self._n), lower=True
        )

        # Scale and shift
        # TODO: can we avoid todense here and just return operator samples?
        if isinstance(self.mean(), scipy.sparse.linalg.LinearOperator):
            mean = self.mean().todense()
        else:
            mean = self.mean()

        # Appendix E: Bartels, S., Probabilistic Linear Algebra, PhD Thesis 2019
        samples_scaled = linops.Symmetrize(dim=self._n) @ (
            linops.Kronecker(A=cholA, B=cholA) @ stdnormal_samples
        )

        return mean[None, :, :] + samples_scaled.T.reshape(-1, self._n, self._n)


def _check_shapes_if_symmetric_kronecker(mean, cov):
    """
    Mean has to be square. If mean has dimension (n x n) then kernels
    factors must be (n x n).
    """
    m, n = mean.shape
    if m != n or n != cov.A.shape[0] or n != cov.B.shape[1]:
        raise ValueError(
            "Normal distributions with symmetric "
            "Kronecker structured kernels must "
            "have square mean and square "
            "kernels factors with matching "
            "dimensions."
        )
